import logging

from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler
from vortex.sqla_orm.OrmCrudHandler import OrmCrudHandler, OrmCrudHandlerExtension

from peek_plugin_diagram._private.PluginNames import diagramFilt

logger = logging.getLogger(__name__)

# This dict matches the definition in the Admin angular app.
filtKey = {"key": "admin.Edit.StringIntTuple"}
filtKey.update(diagramFilt)


# This is the CRUD hander
class __CrudHandler(OrmCrudHandler):
    pass


class __ExtUpdateObservable(OrmCrudHandlerExtension):
    """ Update Observable ORM Crud Extension

    This extension is called after events that will alter data,
    it then notifies the observer.

    """

    def __init__(self, tupleDataObserver: TupleDataObservableHandler) -> None:
        self._tupleDataObserver = tupleDataObserver

    def _tellObserver(self, tuple_, tuples, session, payloadFilt):
        selector = {}
        # Copy any filter values into the selector
        # selector["lookupName"] = payloadFilt["lookupName"]
        tupleSelector = TupleSelector(StringIntTuple.tupleName(),
                                      selector)
        self._tupleDataObserver.notifyOfTupleUpdate(tupleSelector)
        return True

    afterUpdateCommit = _tellObserver
    afterDeleteCommit = _tellObserver


# This method creates an instance of the handler class.
def makeStringIntTableHandler(tupleObservable, dbSessionCreator):
    handler = __CrudHandler(dbSessionCreator, StringIntTuple,
                            filtKey, retreiveAll=True)

    logger.debug("Started")
    handler.addExtension(StringIntTuple, __ExtUpdateObservable(tupleObservable))
    return handler



import logging
from datetime import datetime
from typing import List

import pytz
from sqlalchemy.sql.expression import asc
from twisted.internet import task
from twisted.internet.defer import inlineCallbacks

from peek_plugin_diagram._private.server.controller.StatusController import \
    StatusController
from peek_plugin_diagram._private.storage.GridKeyIndex import \
    DispIndexerQueue as DispIndexerQueueTable
from vortex.DeferUtil import deferToThreadWrapWithLogger, vortexLogFailure

logger = logging.getLogger(__name__)


class DispCompilerQueueController:
    """ Grid Compiler

    Compile the disp items into the grid data

    1) Query for queue
    2) Process queue
    3) Delete from queue
    """

    FETCH_SIZE = 500
    PERIOD = 0.200

    def __init__(self, ormSessionCreator, statusController: StatusController):
        self._ormSessionCreator = ormSessionCreator
        self._statusController: StatusController = statusController

        self._pollLoopingCall = task.LoopingCall(self._poll)
        self._lastQueueId = 0
        self._queueCount = 0

    def isBusy(self) -> bool:
        return self._queueCount != 0

    def start(self):
        self._statusController.setDisplayCompilerStatus(True, self._queueCount)
        d = self._pollLoopingCall.start(self.PERIOD, now=False)
        d.addCallbacks(self._timerCallback, self._timerErrback)

    def _timerErrback(self, failure):
        vortexLogFailure(failure, logger)
        self._statusController.setDisplayCompilerStatus(False, self._queueCount)
        self._statusController.setDisplayCompilerError(str(failure.value))

    def _timerCallback(self, _):
        self._statusController.setDisplayCompilerStatus(False, self._queueCount)

    def stop(self):
        if self._pollLoopingCall.running:
            self._pollLoopingCall.stop()

    def shutdown(self):
        self.stop()

    @inlineCallbacks
    def _poll(self):

        queueItems = yield self._grabQueueChunk()

        if not queueItems:
            return

        queueIds = [o.id for o in queueItems]
        dispIds = list(set([o.dispId for o in queueItems]))

        from peek_plugin_diagram._private.worker.tasks.DispCompilerTask import \
            compileDisps

        try:
            d = compileDisps.delay(queueIds, dispIds)
            d.addCallback(self._pollCallback, datetime.now(pytz.utc), len(queueItems))
            d.addErrback(self._pollErrback, datetime.now(pytz.utc), len(queueItems))
        except Exception as e:
            logger.exception(e)
            return

        # Set the watermark
        self._lastQueueId = queueItems[-1].id

        self._queueCount += 1
        self._statusController.setDisplayCompilerStatus(True, self._queueCount)

    @deferToThreadWrapWithLogger(logger)
    def _grabQueueChunk(self):
        session = self._ormSessionCreator()
        try:
            queueItems = (session.query(DispIndexerQueueTable)
                .order_by(asc(DispIndexerQueueTable.id))
                .filter(DispIndexerQueueTable.id > self._lastQueueId)
                .yield_per(self.FETCH_SIZE)
                .limit(self.FETCH_SIZE)
                .all())

            session.expunge_all()
            return queueItems
        finally:
            session.close()

    @deferToThreadWrapWithLogger(logger)
    def _pollCallback(self, arg, startTime, dispCount):
        self._queueCount -= 1
        logger.debug("%s Disps, Time Taken = %s",
                     dispCount, datetime.now(pytz.utc) - startTime)
        self._statusController.setDisplayCompilerStatus(True, self._queueCount)
        self._statusController.addToDisplayCompilerTotal(self.FETCH_SIZE)

    def _pollErrback(self, failure, startTime, dispCount):
        self._queueCount -= 1
        logger.debug("%s Disps, Time Taken = %s",
                     dispCount, datetime.now(pytz.utc) - startTime)
        self._statusController.setDisplayCompilerStatus(True, self._queueCount)
        self._statusController.setDisplayCompilerError(str(failure.value))
        vortexLogFailure(failure, logger)

    @deferToThreadWrapWithLogger(logger)
    def queueDisps(self, dispIds):
        return self.queueDispIdsToCompile(dispIds, self._ormSessionCreator)

    @classmethod
    def queueDispIdsToCompile(cls, dispIdsToCompile: List[int], ormSessionCreator):
        if not dispIdsToCompile:
            return

        ormSession = ormSessionCreator()
        try:
            cls.queueDispIdsToCompileWithSession(dispIdsToCompile, ormSession)
            ormSession.commit()
        finally:
            ormSession.close()

    @staticmethod
    def queueDispIdsToCompileWithSession(dispIdsToCompile: List[int], ormSessionOrConn):
        if not dispIdsToCompile:
            return

        logger.debug("Queueing %s disps for compile", len(dispIdsToCompile))

        inserts = []
        for dispId in dispIdsToCompile:
            inserts.append(dict(dispId=dispId))

        ormSessionOrConn.execute(DispIndexerQueueTable.__table__.insert(), inserts)



import logging
from datetime import datetime
from typing import List, Callable

import pytz
from sqlalchemy import asc
from twisted.internet import task
from twisted.internet.defer import inlineCallbacks

from peek_plugin_base.storage.StorageUtil import makeCoreValuesSubqueryCondition
from peek_plugin_diagram._private.server.client_handlers.ClientLocationIndexUpdateHandler import \
    ClientLocationIndexUpdateHandler
from peek_plugin_diagram._private.server.controller.StatusController import \
    StatusController
from peek_plugin_diagram._private.storage.LocationIndex import \
    LocationIndexCompilerQueue
from vortex.DeferUtil import deferToThreadWrapWithLogger, vortexLogFailure

logger = logging.getLogger(__name__)


class DispKeyCompilerQueueController:
    """ Disp Compiler

    Compile the disp items into the grid data

    1) Query for queue
    2) Process queue
    3) Delete from queue

    """

    FETCH_SIZE = 10
    PERIOD = 5.000

    QUEUE_MAX = 10
    QUEUE_MIN = 0

    def __init__(self, ormSessionCreator,
                 statusController: StatusController,
                 clientLocationUpdateHandler: ClientLocationIndexUpdateHandler,
                 readyLambdaFunc: Callable):
        self._ormSessionCreator = ormSessionCreator
        self._statusController: StatusController = statusController
        self._clientLocationUpdateHandler: ClientLocationIndexUpdateHandler = clientLocationUpdateHandler
        self._readyLambdaFunc = readyLambdaFunc

        self._pollLoopingCall = task.LoopingCall(self._poll)
        self._lastQueueId = -1
        self._queueCount = 0

    def start(self):
        self._statusController.setLocationIndexCompilerStatus(True, self._queueCount)
        d = self._pollLoopingCall.start(self.PERIOD, now=False)
        d.addCallbacks(self._timerCallback, self._timerErrback)

    def _timerErrback(self, failure):
        vortexLogFailure(failure, logger)
        self._statusController.setLocationIndexCompilerStatus(False, self._queueCount)
        self._statusController.setLocationIndexCompilerError(str(failure.value))

    def _timerCallback(self, _):
        self._statusController.setLocationIndexCompilerStatus(False, self._queueCount)

    def stop(self):
        if self._pollLoopingCall.running:
            self._pollLoopingCall.stop()

    def shutdown(self):
        self.stop()

    @inlineCallbacks
    def _poll(self):
        if not self._readyLambdaFunc():
            return

        from peek_plugin_diagram._private.worker.tasks.LocationIndexCompilerTask import \
            compileLocationIndex

        # We queue the grids in bursts, reducing the work we have to do.
        if self._queueCount > self.QUEUE_MIN:
            return

        # Check for queued items
        queueItems = yield self._grabQueueChunk()
        if not queueItems:
            return

        # De duplicated queued grid keys
        # This is the reason why we don't just queue all the celery tasks in one go.
        # If we keep them in the DB queue, we can remove the duplicates
        # and there are lots of them
        queueIdsToDelete = []

        locationIndexBucketSet = set()
        for i in queueItems:
            if i.indexBucket in locationIndexBucketSet:
                queueIdsToDelete.append(i.id)
            else:
                locationIndexBucketSet.add(i.indexBucket)

        if queueIdsToDelete:
            # Delete the duplicates and requery for our new list
            yield self._deleteDuplicateQueueItems(queueIdsToDelete)
            queueItems = yield self._grabQueueChunk()

        # Send the tasks to the peek worker
        for start in range(0, len(queueItems), self.FETCH_SIZE):

            items = queueItems[start: start + self.FETCH_SIZE]

            # Set the watermark
            self._lastQueueId = items[-1].id

            d = compileLocationIndex.delay(items)
            d.addCallback(self._pollCallback, datetime.now(pytz.utc), len(items))
            d.addErrback(self._pollErrback, datetime.now(pytz.utc))

            self._queueCount += 1
            if self._queueCount >= self.QUEUE_MAX:
                break

    @deferToThreadWrapWithLogger(logger)
    def _grabQueueChunk(self):
        session = self._ormSessionCreator()
        try:
            qry = (session.query(LocationIndexCompilerQueue)
                .order_by(asc(LocationIndexCompilerQueue.id))
                .filter(LocationIndexCompilerQueue.id > self._lastQueueId)
                .yield_per(500)
                # .limit(self.FETCH_SIZE)
                )

            queueItems = qry.all()
            session.expunge_all()

            return queueItems

        finally:
            session.close()

    @deferToThreadWrapWithLogger(logger)
    def _deleteDuplicateQueueItems(self, itemIds):
        session = self._ormSessionCreator()
        table = LocationIndexCompilerQueue.__table__
        try:
            SIZE = 1000
            for start in range(0, len(itemIds), SIZE):
                chunkIds = itemIds[start: start + SIZE]

                session.execute(table.delete(makeCoreValuesSubqueryCondition(
                    session.bind, table.c.id, chunkIds
                )))

            session.commit()
        finally:
            session.close()

    def _pollCallback(self, indexBuckets: List[str], startTime, processedCount):
        self._queueCount -= 1
        logger.debug("Time Taken = %s" % (datetime.now(pytz.utc) - startTime))
        self._clientLocationUpdateHandler.sendLocationIndexes(indexBuckets)
        self._statusController.addToLocationIndexCompilerTotal(processedCount)
        self._statusController.setLocationIndexCompilerStatus(True, self._queueCount)

    def _pollErrback(self, failure, startTime):
        self._queueCount -= 1
        self._statusController.setLocationIndexCompilerError(str(failure.value))
        self._statusController.setLocationIndexCompilerStatus(True, self._queueCount)
        logger.debug("Time Taken = %s" % (datetime.now(pytz.utc) - startTime))
        vortexLogFailure(failure, logger)



import logging

from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer, String

from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from vortex.Tuple import Tuple, addTupleType
from peek_plugin_diagram._private.storage.DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)


@addTupleType
class BranchIndexCompilerQueue(Tuple, DeclarativeBase):
    __tablename__ = 'BranchIndexCompilerQueue'
    __tupleType__ = diagramTuplePrefix + 'BranchIndexCompilerQueueTable'

    id = Column(Integer, primary_key=True, autoincrement=True)

    modelSetId = Column(Integer,
                        ForeignKey('ModelSet.id', ondelete='CASCADE'),
                        nullable=False,
                        autoincrement=True)

    chunkKey = Column(String, primary_key=True)



import logging
from collections import defaultdict
from datetime import datetime
from typing import List, Dict

from twisted.internet.defer import DeferredList, Deferred, inlineCallbacks

from peek_plugin_diagram._private.PluginNames import diagramFilt
from peek_plugin_diagram._private.client.controller.GridCacheController import \
    GridCacheController
from peek_plugin_diagram._private.server.client_handlers.ClientGridLoaderRpc import \
    ClientGridLoaderRpc
from vortex.DeferUtil import vortexLogFailure
from vortex.Payload import Payload
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.VortexABC import SendVortexMsgResponseCallable
from vortex.VortexFactory import VortexFactory

logger = logging.getLogger(__name__)

clientGridWatchUpdateFromDeviceFilt = {'key': "clientGridWatchUpdateFromDevice"}
clientGridWatchUpdateFromDeviceFilt.update(diagramFilt)

#: This the type of the data that we get when the clients observe new grids.
DeviceGridT = Dict[str, datetime]


# ModelSet HANDLER
class GridCacheHandler(object):
    def __init__(self, gridCacheController: GridCacheController, clientId: str):
        """ App Grid Handler

        This class handles the custom needs of the desktop/mobile apps observing grids.

        """
        self._gridCacheController = gridCacheController
        self._clientId = clientId

        self._epObserve = PayloadEndpoint(
            clientGridWatchUpdateFromDeviceFilt, self._processObserve
        )

        self._observedGridKeysByVortexUuid = defaultdict(list)
        self._observedVortexUuidsByGridKey = defaultdict(list)

    def shutdown(self):
        self._epObserve.shutdown()
        self._epObserve = None

    # ---------------
    # Filter out offline vortexes

    def _filterOutOfflineVortexes(self):
        # TODO, Change this to observe offline vortexes
        # This depends on the VortexFactory offline observable implementation.
        # Which is incomplete at this point :-|

        vortexUuids = set(VortexFactory.getRemoteVortexUuids())
        vortexUuidsToRemove = set(self._observedGridKeysByVortexUuid) - vortexUuids

        if not vortexUuidsToRemove:
            return

        for vortexUuid in vortexUuidsToRemove:
            del self._observedGridKeysByVortexUuid[vortexUuid]

        self._rebuildStructs()

    # ---------------
    # Process update from the server

    def notifyOfGridUpdate(self, gridKeys: List[str]):
        """ Notify of Grid Updates

        This method is called by the client.GridCacheController when it receives updates
        from the server.

        """
        self._filterOutOfflineVortexes()

        payloadsByVortexUuid = defaultdict(Payload)

        for gridKey in gridKeys:

            gridTuple = self._gridCacheController.grid(gridKey)
            vortexUuids = self._observedVortexUuidsByGridKey.get(gridKey, [])

            # Queue up the required client notifications
            for vortexUuid in vortexUuids:
                logger.debug("Sending unsolicited grid %s to vortex %s",
                             gridKey, vortexUuid)
                payloadsByVortexUuid[vortexUuid].tuples.append(gridTuple)

        # Send the updates to the clients
        dl = []
        for vortexUuid, payload in list(payloadsByVortexUuid.items()):
            payload.filt = clientGridWatchUpdateFromDeviceFilt

            # Serliase in thread, and then send.
            d = payload.makePayloadEnvelopeDefer()
            d.addCallback(lambda payloadEnvelope:payloadEnvelope.toVortexMsgDefer())
            d.addCallback(VortexFactory.sendVortexMsg, destVortexUuid=vortexUuid)
            dl.append(d)

        # Log the errors, otherwise we don't care about them
        dl = DeferredList(dl, fireOnOneErrback=True)
        dl.addErrback(vortexLogFailure, logger, consumeError=True)

    # ---------------
    # Process observes from the devices

    @inlineCallbacks
    def _processObserve(self, payloadEnvelope: PayloadEnvelope,
                        vortexUuid: str,
                        sendResponse: SendVortexMsgResponseCallable,
                        **kwargs):
        cacheAll = payloadEnvelope.filt.get("cacheAll") == True

        payload = yield payloadEnvelope.decodePayloadDefer()

        lastUpdateByGridKey: DeviceGridT = payload.tuples[0]

        if not cacheAll:
            gridKeys = list(lastUpdateByGridKey.keys())
            self._observedGridKeysByVortexUuid[vortexUuid] = gridKeys
            self._rebuildStructs()

        self._replyToObserve(payload.filt,
                             lastUpdateByGridKey,
                             sendResponse,
                             cacheAll=cacheAll)

    def _rebuildStructs(self) -> None:
        """ Rebuild Structs

        Rebuild the reverse index of uuids by grid key.

        :returns: None
        """
        # Rebuild the other reverse lookup
        newDict = defaultdict(list)

        for vortexUuid, gridKeys in self._observedGridKeysByVortexUuid.items():
            for gridKey in gridKeys:
                newDict[gridKey].append(vortexUuid)

        keysChanged = set(self._observedVortexUuidsByGridKey) != set(newDict)

        self._observedVortexUuidsByGridKey = newDict

        # Notify the server that this client service is watching different grids.
        if keysChanged:
            d = ClientGridLoaderRpc.updateClientWatchedGrids(
                clientId=self._clientId,
                gridKeys=list(self._observedVortexUuidsByGridKey)
            )
            d.addErrback(vortexLogFailure, logger, consumeError=False)

    # ---------------
    # Reply to device observe

    def _replyToObserve(self, filt,
                        lastUpdateByGridKey: DeviceGridT,
                        sendResponse: SendVortexMsgResponseCallable,
                        cacheAll=False) -> None:
        """ Reply to Observe

        The client has told us that it's observing a new set of grids, and the lastUpdate
        it has for each of those grids. We will send them the grids that are out of date
        or missing.

        :param filt: The payload filter to respond to.
        :param lastUpdateByGridKey: The dict of gridKey:lastUpdate
        :param sendResponse: The callable provided by the Vortex (handy)
        :returns: None

        """
        gridTuplesToSend = []

        def sendChunk(toSend):
            if not toSend and not cacheAll:
                return

            payload = Payload(filt=filt, tuples=toSend)
            d: Deferred = payload.makePayloadEnvelopeDefer(compressionLevel=2)
            d.addCallback(lambda payloadEnvelope: payloadEnvelope.toVortexMsgDefer())
            d.addCallback(sendResponse)
            d.addErrback(vortexLogFailure, logger, consumeError=True)

        # Check and send any updates
        for gridKey, lastUpdate in lastUpdateByGridKey.items():
            # NOTE: lastUpdate can be null.
            gridTuple = self._gridCacheController.grid(gridKey)
            if not gridTuple:
                logger.debug("Grid %s is not in the cache" % gridKey)
                continue

            # We are king, If it's it's not our version, it's the wrong version ;-)
            logger.debug("%s, %s,  %s", gridTuple.lastUpdate == lastUpdate,
                         gridTuple.lastUpdate, lastUpdate)
            if gridTuple.lastUpdate == lastUpdate:
                logger.debug("Grid %s matches the cache" % gridKey)
            else:
                gridTuplesToSend.append(gridTuple)
                logger.debug("Sending grid %s from the cache" % gridKey)

            if len(gridTuplesToSend) == 5 and not cacheAll:
                sendChunk(gridTuplesToSend)
                gridTuplesToSend = []

        sendChunk(gridTuplesToSend)



from typing import Optional, List

from geoalchemy2 import WKBElement
from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from peek_plugin_diagram.tuples.model.ImportLiveDbDispLinkTuple import \
    ImportLiveDbDispLinkTuple
from vortex.Tuple import Tuple, addTupleType, TupleField


@addTupleType
class ImportDispPolygonTuple(Tuple):
    """ Imported Display Polygon

    This tuple is used by other plugins to load polygons into the diagram.

    """
    __tupleType__ = diagramTuplePrefix + 'ImportDispPolygonTuple'

    ### BEGIN DISP COMMON FIELDS ###

    #: Key, This value is a unique ID of the object that this graphic represents
    # It's used to link this graphical object to objects in other plugins, like vertices
    # in the peek-plugin-graphdb plugin.
    # Length = 50
    key: str = TupleField()

    #: Selectable, Is is this item selectable?, the layer also needs selectable=true
    selectable: bool = TupleField()

    #: Data, Generic data, this is passed to the popup context in the UI.
    # peek_plugin_diagram doesn't care as long as it's json compatible or None
    # Json length Length = 400
    data: Optional[dict] = TupleField(None)

    #: The hash of the level to link to (Matches ImportDispLevel.importHash)
    levelHash: str = TupleField()

    #: The hash of the layer to link to (Matches ImportDispLayer.importHash)
    layerHash: str = TupleField()

    #: The Z Order of this display object when compared against other objects on
    # same layer and level.
    zOrder: int = TupleField()

    #: The unique hash of this display object
    importHash: str = TupleField()

    #: The unique hash for all the display items imported in a group with this one.
    #: for example, a page or tile reference.
    importGroupHash: str = TupleField()

    #: The key of the ModelSet to import into
    modelSetKey: str = TupleField()

    #: The Key of the Coordinate Set to import into
    coordSetKey: str = TupleField()

    #: Related links to LiveDB values for this display item
    liveDbDispLinks: List[ImportLiveDbDispLinkTuple] = TupleField()

    #: Parent DispGroup Hash, If this disp is part of a disp group then set this field to
    # the ImportDispGroupTuple.importHash fields value
    # NOTE: If this disp is part of a display group, then the GEOM coordinates need to
    # be relative to 0x0.
    # NOTE: Disps that are apart of a group must all be imported with the same
    # importGroupHash, during the same import call.
    parentDispGroupHash: str = TupleField()

    ### BEGIN FIELDS FOR THIS DISP ###

    lineWidth: int = TupleField()
    lineStyleHash: str = TupleField()
    lineColorHash: Optional[str] = TupleField()
    fillColorHash: Optional[str] = TupleField()

    FILL_TOP_TO_BOTTOM = 0
    FILL_BOTTOM_TO_TOP = 1
    FILL_RIGHT_TO_LEFT = 2
    FILL_LEFT_TO_RIGHT = 3
    fillDirection: int = TupleField()
    fillPercent: int = TupleField()

    geom: WKBElement = TupleField()

    #: Is this a retangle, with perpendicular equal sides.
    # This will effect how the polygon is edited
    isRectangle: Optional[bool] = TupleField()



from datetime import datetime

from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from vortex.Tuple import addTupleType, TupleField, Tuple


@addTupleType
class EncodedGridTuple(Tuple):
    """ Encoded Grid Tuple

    This tuple stores a pre-encoded version of a GridTuple

    """
    __tupleType__ = diagramTuplePrefix + "EncodedGridTuple"

    gridKey: str = TupleField()

    # A GridTuple, already encoded and ready for storage in the clients cache
    encodedGridTuple: str = TupleField()

    lastUpdate: datetime = TupleField()


"""added location index

Peek Plugin Database Migration Script

Revision ID: 23117803d414
Revises: 1fe753e1f369
Create Date: 2017-09-09 16:10:10.582473

"""

# revision identifiers, used by Alembic.
revision = '23117803d414'
down_revision = '1fe753e1f369'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('LocationIndexCompiled',
    sa.Column('id', sa.Integer(), nullable=False, autoincrement=True),
    sa.Column('indexBucket', sa.String(length=100), nullable=False),
    sa.Column('blobData', sa.LargeBinary(), nullable=False),
    sa.Column('lastUpdate', sa.String(length=50), nullable=False),
    sa.Column('modelSetId', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['modelSetId'], ['pl_diagram.ModelSet.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id', 'indexBucket'),
    schema='pl_diagram'
    )
    op.create_index('idx_DKIndexUpdate_coordSetId', 'LocationIndexCompiled', ['modelSetId'], unique=False, schema='pl_diagram')
    op.create_index('idx_DKIndexUpdate_gridKey', 'LocationIndexCompiled', ['indexBucket'], unique=True, schema='pl_diagram')
    op.create_table('LocationIndexCompilerQueue',
    sa.Column('id', sa.Integer(), nullable=False, autoincrement=True),
    sa.Column('indexBucket', sa.String(length=100), nullable=False),
    sa.Column('modelSetId', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['modelSetId'], ['pl_diagram.ModelSet.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id', 'indexBucket', 'modelSetId'),
    schema='pl_diagram'
    )
    op.create_index('idx_DKCompQueue_coordSetId_gridKey', 'LocationIndexCompilerQueue', ['modelSetId', 'indexBucket'], unique=False, schema='pl_diagram')
    op.create_table('LocationIndex',
    sa.Column('indexBucket', sa.String(length=100), nullable=False),
    sa.Column('dispId', sa.Integer(), nullable=False),
    sa.Column('modelSetId', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['dispId'], ['pl_diagram.DispBase.id'], ondelete='CASCADE'),
    sa.ForeignKeyConstraint(['modelSetId'], ['pl_diagram.ModelSet.id'], ),
    sa.PrimaryKeyConstraint('indexBucket', 'dispId'),
    schema='pl_diagram'
    )
    op.create_index('idx_LocationIndex_dispId', 'LocationIndex', ['dispId'], unique=False, schema='pl_diagram')
    op.create_index('idx_LocationIndex_indexBucket', 'LocationIndex', ['indexBucket'], unique=False, schema='pl_diagram')
    op.create_index('idx_LocationIndex_modelSetId', 'LocationIndex', ['modelSetId'], unique=False, schema='pl_diagram')
    op.add_column('DispBase', sa.Column('locationJson', sa.String(length=120), nullable=True), schema='pl_diagram')



    # ### end Alembic commands ###


def downgrade():

    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('DispBase', 'locationJson', schema='pl_diagram')
    op.drop_index('idx_LocationIndex_modelSetId', table_name='LocationIndex', schema='pl_diagram')
    op.drop_index('idx_LocationIndex_indexBucket', table_name='LocationIndex', schema='pl_diagram')
    op.drop_index('idx_LocationIndex_dispId', table_name='LocationIndex', schema='pl_diagram')
    op.drop_table('LocationIndex', schema='pl_diagram')
    op.drop_index('idx_DKCompQueue_coordSetId_gridKey', table_name='LocationIndexCompilerQueue', schema='pl_diagram')
    op.drop_table('LocationIndexCompilerQueue', schema='pl_diagram')
    op.drop_index('idx_DKIndexUpdate_gridKey', table_name='LocationIndexCompiled', schema='pl_diagram')
    op.drop_index('idx_DKIndexUpdate_coordSetId', table_name='LocationIndexCompiled', schema='pl_diagram')
    op.drop_table('LocationIndexCompiled', schema='pl_diagram')
    # ### end Alembic command


import logging
from peek_plugin_base.worker.PluginWorkerEntryHookABC import PluginWorkerEntryHookABC

from peek_plugin_diagram._private.worker.tasks import GridCompilerTask, \
    ImportDispTask, LiveDbDisplayValueConverterTask, DispCompilerTask, \
    LocationIndexCompilerTask
from peek_plugin_diagram._private.worker.tasks.branch import BranchIndexCompiler, \
    BranchIndexImporter, BranchIndexUpdater

logger = logging.getLogger(__name__)


class WorkerEntryHook(PluginWorkerEntryHookABC):
    def load(self):
        logger.debug("loaded")

    def start(self):
        logger.debug("started")

    def stop(self):
        logger.debug("stopped")

    def unload(self):
        logger.debug("unloaded")

    @property
    def celeryAppIncludes(self):
        return [BranchIndexUpdater.__name__,
                BranchIndexCompiler.__name__,
                BranchIndexImporter.__name__,
                DispCompilerTask.__name__,
                GridCompilerTask.__name__,
                LiveDbDisplayValueConverterTask.__name__,
                ImportDispTask.__name__,
                LocationIndexCompilerTask.__name__]

    @property
    def celeryApp(self):
        from .CeleryApp import celeryApp
        return celeryApp



import logging

from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from sqlalchemy import Column, LargeBinary
from sqlalchemy import ForeignKey
from sqlalchemy import Integer, String
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import Index

from peek_plugin_diagram._private.storage.ModelSet import ModelSet
from vortex.Tuple import Tuple, addTupleType
from peek_plugin_diagram._private.storage.DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)



@addTupleType
class BranchIndexEncodedChunk(Tuple, DeclarativeBase):
    __tablename__ = 'BranchIndexEncodedChunk'
    __tupleType__ = diagramTuplePrefix + 'BranchIndexEncodedChunkTable'

    id = Column(Integer, primary_key=True, autoincrement=True)

    modelSetId = Column(Integer,
                        ForeignKey('ModelSet.id', ondelete='CASCADE'),
                        nullable=False)
    modelSet = relationship(ModelSet)

    chunkKey = Column(String, nullable=False)
    encodedData = Column(LargeBinary, nullable=False)
    encodedHash = Column(String, nullable=False)
    lastUpdate = Column(String, nullable=False)

    __table_args__ = (
        Index("idx_BranchIndexEnc_modelSetId_chunkKey", modelSetId, chunkKey, unique=False),
    )



import ujson as json
import logging
from typing import List

from peek_plugin_diagram._private.worker.tasks.ImportDispTask import importDispsTask
from peek_plugin_livedb.server.LiveDBWriteApiABC import LiveDBWriteApiABC
from twisted.internet.defer import inlineCallbacks
from vortex.SerialiseUtil import convertFromWkbElement

logger = logging.getLogger(__name__)


class DispImportController:
    def __init__(self, liveDbWriteApi: LiveDBWriteApiABC):
        self._liveDbWriteApi = liveDbWriteApi

    def shutdown(self):
        self._liveDbWriteApi = None

    @inlineCallbacks
    def importDisps(self, modelSetKey: str, coordSetKey: str,
                    importGroupHash: str, dispsEncodedPayload: bytes):

        liveDbItemsToImport = yield importDispsTask.delay(
            modelSetKey, coordSetKey, importGroupHash, dispsEncodedPayload
        )

        if liveDbItemsToImport:
            yield self._liveDbWriteApi.importLiveDbItems(
                modelSetKey, liveDbItemsToImport
            )

            yield self._liveDbWriteApi.pollLiveDbValueAcquisition(
                modelSetKey, [i.key for i in liveDbItemsToImport]
            )



import logging
from _collections import defaultdict
from collections import namedtuple
from datetime import datetime
from typing import List, Dict

import pytz
import ujson as json
from peek_plugin_base.worker import CeleryDbConn
from peek_plugin_diagram._private.storage.Display import DispBase, DispTextStyle, \
    DispGroup, DispGroupPointer
from peek_plugin_diagram._private.storage.GridKeyIndex import GridKeyIndex, \
    DispIndexerQueue, GridKeyCompilerQueue
from peek_plugin_diagram._private.storage.LiveDbDispLink import LiveDbDispLink
from peek_plugin_diagram._private.storage.LocationIndex import LocationIndex, \
    LocationIndexCompilerQueue
from peek_plugin_diagram._private.storage.ModelSet import ModelCoordSet, \
    makeDispGroupGridKey
from peek_plugin_diagram._private.worker.CeleryApp import celeryApp
from peek_plugin_diagram._private.worker.tasks._CalcDisp import \
    _scaleDispGeomWithCoordSet, _scaleDispGeom, _createHashId
from peek_plugin_diagram._private.worker.tasks._CalcDispFromLiveDb import \
    _mergeInLiveDbValues
from peek_plugin_diagram._private.worker.tasks._CalcGridForDisp import makeGridKeysForDisp
from peek_plugin_diagram._private.worker.tasks._CalcLocation import makeLocationJson, \
    dispKeyHashBucket
from peek_plugin_livedb.worker.WorkerApi import WorkerApi
from sqlalchemy import select, join
from sqlalchemy.orm import subqueryload
from sqlalchemy.sql.selectable import Select
from txcelery.defer import DeferrableTask

logger = logging.getLogger(__name__)

CoordSetIdGridKeyData = namedtuple("CoordSetIdGridKeyTuple",
                                   ["coordSetId", "gridKey"])

ModelSetIdIndexBucketData = namedtuple("ModelSetIdIndexBucketTuple",
                                       ["modelSetId", "indexBucket"])

DispData = namedtuple('DispData', ['json', 'levelOrder', 'layerOrder'])

PreparedDisp = namedtuple('PreparedDisp', ['disp', 'geomArray', 'dispDict'])


@DeferrableTask
@celeryApp.task(bind=True)
def compileDisps(self, queueIds, dispIds):
    """ Compile Disps

    This function takes a list of Disp IDs and compiles them.
    The processing is as follows (more or less)

    0) Load lookups

    ----

    1) DispGroupPointers, copy disps from group to pointer

    ----

    2) Load the Disps from the DB

    3) Apply the LiveDB values to the Disp attributes

    4) Scale the Disp geomJson to match the coord set scaling

    5) DispGroups, take Disps as part of a disp group and load them into JSON in the
        DispGroup. PreparedDisp????

    6) Extract any new LocationIndex entries, of the Disp has a key

    7) Determine which grids this disp will live in, and create GridKeyIndex entries
        for those grid keys for this disp.

    8) Write the Disp JSON back to the disp

    ormSession.commit() here.
        This stores the following updates that have been made into the disp:
        * dispJson,
        * locationJson,
        * livedb attribute updates

    ----

    9) Write the calculated data to tables

    NOTE: Disps that belong to a DispGroup will not be queued for compile by
    ImportDispTask.

    """

    # ==========================
    # 0) Load the lookups
    ormSession = CeleryDbConn.getDbSession()
    try:
        # ---------------
        # Load Coord Sets
        coordSets = (ormSession.query(ModelCoordSet)
                     .options(subqueryload(ModelCoordSet.modelSet),
                              subqueryload(ModelCoordSet.gridSizes))
                     .all())

        # Get Model Set Name Map
        coordSetById = {o.id: o for o in coordSets}

        # ---------------
        # Load Coord Sets
        textStyleById = {ts.id: ts for ts in ormSession.query(DispTextStyle).all()}

        ormSession.expunge_all()

    except Exception as e:
        logger.exception(e)
        raise self.retry(exc=e, countdown=2)

    finally:
        ormSession.close()

    # ==========================
    # This method will create new disps that will be compiled later.
    try:

        # ---------------
        # 1) Clone the disps for the group instances
        dispIdsIncludingClones = _cloneDispsForDispGroupPointer(dispIds)

    except Exception as e:
        logger.exception(e)
        raise self.retry(exc=e, countdown=2)

    # ==========================
    # Run all the ORM Session update methods
    ormSession = CeleryDbConn.getDbSession()
    try:
        with ormSession.no_autoflush:
            # ---------------
            # 2) Apply the LiveDB Attribute updates
            disps = _loadDisps(ormSession, dispIdsIncludingClones)

            # ---------------
            # 3) Apply the LiveDB Attribute updates
            _applyLiveDbAttributes(ormSession, disps, coordSetById)

            # ---------------
            # 4) Scale the Disp geomJson to match the coord set scaling
            preparedDisps = _scaleDisp(disps, coordSetById)

            # 5) DispGroups, take Disps as part of a disp group and load them
            # into JSON in the DispGroup. PreparedDisp????
            _compileDispGroups(ormSession, preparedDisps)

            # ---------------
            # 6) Extract any new LocationIndex entries, of the Disp has a key
            locationCompiledQueueItems, locationIndexByDispId = _indexLocation(
                preparedDisps, coordSetById
            )

            # ---------------
            # 7) Determine which grids this disp will live in, and create GridKeyIndex
            # entries for those grid keys for this disp.
            gridCompiledQueueItems, gridKeyIndexesByDispId = _calculateGridKeys(
                preparedDisps, coordSetById, textStyleById
            )

            # ---------------
            # 8) Write the Disp JSON back to the disp
            _updateDispsJson(preparedDisps)

        # ---------------
        # Commit the updates
        startTime = datetime.now(pytz.utc)
        ormSession.commit()
        logger.debug("Committed %s disp objects in %s",
                     len(disps), (datetime.now(pytz.utc) - startTime))

    except Exception as e:
        ormSession.rollback()
        logger.exception(e)
        raise self.retry(exc=e, countdown=2)

    finally:
        ormSession.close()

    # ==========================
    # 9) Run the bulk DB delete/insert methods
    try:

        _insertToDb(dispIdsIncludingClones,
                    gridCompiledQueueItems, gridKeyIndexesByDispId,
                    locationCompiledQueueItems, locationIndexByDispId, queueIds)

    except Exception as e:
        logger.exception(e)
        raise self.retry(exc=e, countdown=2)

    logger.info("Compiled %s disp objects in %s",
                len(dispIds), (datetime.now(pytz.utc) - startTime))


def _queryDispsForGroup(ormSession, dispGroupIds) -> Dict[int, List]:
    qry = ormSession.query(DispBase) \
        .options(subqueryload(DispBase.liveDbLinks)) \
        .filter(DispBase.groupId.in_(dispGroupIds))

    dispsByGroupId = defaultdict(list)
    for o in qry.all():
        dispsByGroupId[o.groupId].append(o)

    return dispsByGroupId


def _cloneDispsForDispGroupPointer(dispIds: List[int]):
    """ Clone Disps for DispGroupPointer

    This method will clone "instances" of the disps in the disp groups for the
    DispGroupPointer.


    """
    startTime = datetime.now(pytz.utc)

    ormSession = CeleryDbConn.getDbSession()
    try:

        # -----
        # Load the disp group pointers
        qry = ormSession.query(DispGroupPointer) \
            .filter(DispGroupPointer.targetDispGroupId != None) \
            .filter(DispGroupPointer.id.in_(dispIds))

        dispGroupPointers: List[DispGroupPointer] = qry.all()

        # If there are no DispGroupPointers that need cloning, then return.
        if not dispGroupPointers:
            logger.debug("Cloning skipped,"
                         " there are no disp group ptrs with targets, in %s",
                         (datetime.now(pytz.utc) - startTime))
            return dispIds

        dispGroupPointerTargetIds = [o.targetDispGroupId for o in dispGroupPointers]

        del qry

        # -----
        # Delete any existing disps are in these pointers
        ormSession.query(DispBase) \
            .filter(DispBase.groupId.in_([o.id for o in dispGroupPointers])) \
            .delete(synchronize_session=False)

        ormSession.commit()

        # -----
        # Query for the disp groups we'll need
        dispGroupChildsByGroupId = _queryDispsForGroup(ormSession,
                                                       dispGroupPointerTargetIds)

        # -----
        # Query for the disp groups names
        dispBaseTable = DispBase.__table__
        dispGroupTable = DispGroup.__table__

        qry = ormSession.execute(
            select(columns=[dispBaseTable.c.id,
                            dispBaseTable.c.coordSetId,
                            dispGroupTable.c.name],
                   whereclause=dispBaseTable.c.id.in_(dispGroupPointerTargetIds))
                .select_from(join(dispGroupTable, dispBaseTable,
                                  dispGroupTable.c.id == dispBaseTable.c.id))
        )

        dispGroupNameByGroupId = {o.id: '%s|%s' % (o.coordSetId, o.name)
                                  for o in qry.fetchall()}

        del qry

        # -----
        # Clone the child disps
        cloneDisps = []
        cloneLiveDbDispLinks = []

        for dispPtr in dispGroupPointers:
            if not dispPtr.targetDispGroupId:
                logger.debug("Pointer has no targetGroupId id=%s", dispPtr.id)
                continue

            dispGroupChilds = dispGroupChildsByGroupId.get(dispPtr.targetDispGroupId)

            if not dispGroupChilds:
                logger.warning("Pointer points to missing DispGroup,"
                               " id=%s, targetGroupId=%s", dispPtr.id,
                               dispPtr.targetDispGroupId)
                continue

            x, y = json.loads(dispPtr.geomJson)
            dispPtr.targetDispGroupName = \
                dispGroupNameByGroupId[dispPtr.targetDispGroupId]

            for templateDisp in dispGroupChilds:
                # Create the clone
                cloneDisp = templateDisp.tupleClone()
                cloneDisps.append(cloneDisp)

                cloneDisp.coordSetId = dispPtr.coordSetId

                # Offset the geometry
                geom = json.loads(cloneDisp.geomJson)
                geom = _scaleDispGeom(geom, 1, 1, x, y)
                cloneDisp.geomJson = json.dumps(geom)

                # Assign the clone to the DispGroupPointer
                cloneDisp.groupId = dispPtr.id

                for dispLink in templateDisp.liveDbLinks:
                    cloneDispLink = dispLink.tupleClone()
                    cloneLiveDbDispLinks.append(cloneDispLink)

                    cloneDispLink.id = None
                    cloneDispLink.disp = cloneDisp
                    cloneDispLink.coordSetId = dispPtr.coordSetId

        # -----
        # Preallocate the IDs for performance on PostGreSQL
        dispIdGen = CeleryDbConn.prefetchDeclarativeIds(DispBase, len(cloneDisps))
        for cloneDisp in cloneDisps:
            cloneDisp.id = next(dispIdGen)

        # Preallocate the IDs for performance on PostGreSQL
        dispLinkIdGen = CeleryDbConn.prefetchDeclarativeIds(LiveDbDispLink,
                                                            len(cloneLiveDbDispLinks))
        for cloneDispLink in cloneLiveDbDispLinks:
            cloneDispLink.id = next(dispLinkIdGen)
            cloneDispLink.dispId = cloneDispLink.disp.id
            cloneDispLink.disp = None

        # -----
        # Create the new list of IDs to compile
        # Do this here, otherwise it will cause a DB refresh if it's after the commit.
        dispIdsIncludingClones = dispIds + [o.id for o in cloneDisps]

        ormSession.bulk_save_objects(cloneDisps, update_changed_only=False)
        ormSession.bulk_save_objects(cloneLiveDbDispLinks, update_changed_only=False)

        ormSession.commit()

        logger.debug("Cloned %s disp group objects in %s",
                     len(cloneDisps), (datetime.now(pytz.utc) - startTime))

    except Exception:
        ormSession.rollback()
        raise

    finally:
        ormSession.close()

    return dispIdsIncludingClones


def _loadDisps(ormSession, dispIdsIncludingClones: List[int]):
    """ Load Disps

    This method loads the disps from the database

    """
    startTime = datetime.now(pytz.utc)

    # -----
    # Begin the DISP merge from live data
    qry = (
        ormSession.query(DispBase)
            .options(subqueryload(DispBase.liveDbLinks),
                     subqueryload(DispBase.level))
            .filter(DispBase.id.in_(dispIdsIncludingClones))
    )

    allDisps = qry.all()

    logger.debug("Loaded %s disp objects in %s",
                 len(allDisps), (datetime.now(pytz.utc) - startTime))

    return allDisps


def _applyLiveDbAttributes(ormSession, disps, coordSetById):
    startTime = datetime.now(pytz.utc)

    # Prepare the LiveDbKeys
    liveDbKeysByModelSetKey = defaultdict(list)
    for disp in disps:
        # Add a reference to the model set name for convenience
        disp.modelSetKey = coordSetById[disp.coordSetId].modelSet.key
        liveDbKeysByModelSetKey[disp.modelSetKey].extend(
            [dl.liveDbKey for dl in disp.liveDbLinks]
        )

    liveDbItemByModelSetKeyByKey = {}
    for modelSetKey, liveDbKeys in liveDbKeysByModelSetKey.items():
        liveDbItemByModelSetKeyByKey[modelSetKey] = {
            i.key: i for i in
            WorkerApi.getLiveDbDisplayValues(ormSession, modelSetKey, liveDbKeys)
        }

    logger.debug("Loaded LiveDB Attributes for %s disp objects in %s",
                 len(disps), (datetime.now(pytz.utc) - startTime))

    for disp in disps:  # Work out which grids we belong to
        # Get the LiveDB Items for this model set
        liveDbItemByKey = liveDbItemByModelSetKeyByKey[disp.modelSetKey]

        # Apply live db attributes
        _mergeInLiveDbValues(disp, liveDbItemByKey)

    logger.debug("Applied LiveDB values to %s disp objects in %s",
                 len(disps), (datetime.now(pytz.utc) - startTime))


def _scaleDisp(disps, coordSetById):
    """ Scale Disps

    This method scales the display item geometry as per the coord set

    """

    startTime = datetime.now(pytz.utc)

    # inserts for GridKeyCompilerQueue
    preparedDisps: List[PreparedDisp] = []

    for disp in disps:
        # Get a reference to the coordSet
        coordSet = coordSetById[disp.coordSetId]

        # Create the JSON Dict
        dispDict = disp.tupleToSmallJsonDict()

        # Get and Scale the Geometry
        geomArray = None
        # Disp Groups have no geometry
        if not isinstance(disp, DispGroup):
            geomArray = json.loads(disp.geomJson)
            geomArray = _scaleDispGeomWithCoordSet(geomArray, coordSet)
            dispDict["g"] = geomArray

        preparedDisps.append(PreparedDisp(disp, geomArray, dispDict))

    logger.debug("Scaled %s disps in %s",
                 len(disps), (datetime.now(pytz.utc) - startTime))

    return preparedDisps


def _compileDispGroups(ormSession, preparedDisps: List[PreparedDisp]):
    """ Compile Disp Groups

    This method will pack the child disps into the disp groups dispJson field.

    """

    def packDisp(disp):
        """ Pack Disp

        """
        dispDict = disp.tupleToSmallJsonDict(includeFalse=False, includeNones=False)
        dispDict["g"] = json.loads(disp.geomJson)
        return dispDict

    startTime = datetime.now(pytz.utc)

    preparedDispGroupByIds: Dict[int, PreparedDisp] = {
        o.disp.id: o
        for o in preparedDisps
        if isinstance(o.disp, DispGroup)
    }

    # Query for the disp groups with loaded child disps we'll need

    childDispsByGroupId = _queryDispsForGroup(ormSession, preparedDispGroupByIds)

    childDispCount = 0

    for groupId in preparedDispGroupByIds:
        preparedDispGroup = preparedDispGroupByIds[groupId]
        childDisps = childDispsByGroupId[groupId]

        preparedDispGroup.dispDict['di'] = json.dumps([
            packDisp(disp) for disp in childDisps
        ])

        childDispCount += len(childDisps)

    logger.debug("Packed %s disps into %s group objects in %s",
                 childDispCount, len(preparedDispGroupByIds),
                 (datetime.now(pytz.utc) - startTime))


def _indexLocation(preparedDisps, coordSetById):
    """ Index Location

    This method extracts the location index data from the disps.

    """

    startTime = datetime.now(pytz.utc)

    # ---------------
    # inserts for GridKeyCompilerQueue
    locationCompiledQueueItems = set()

    # List of location index bucket to disp index items to insert into LocationIndex
    locationIndexByDispId = {}

    count = 0

    for pdisp in preparedDisps:
        if not pdisp.geomArray or not pdisp.disp.key:
            continue
        count += 1

        # Get a reference to the coordSet
        coordSet = coordSetById[pdisp.disp.coordSetId]

        # Create the location json for the LocationIndex
        pdisp.disp.locationJson = makeLocationJson(pdisp.disp, pdisp.geomArray)

        # Create the index bucket
        indexBucket = dispKeyHashBucket(coordSet.modelSet.name, pdisp.disp.key)

        # Create the compiler queue item
        locationCompiledQueueItems.add(
            ModelSetIdIndexBucketData(modelSetId=coordSet.modelSetId,
                                      indexBucket=indexBucket)
        )

        # Create the location index item
        locationIndexByDispId[pdisp.disp.id] = dict(
            indexBucket=indexBucket,
            dispId=pdisp.disp.id,
            modelSetId=coordSet.modelSetId
        )

    logger.debug("Indexed %s disp Locations in %s",
                 count, (datetime.now(pytz.utc) - startTime))

    return locationCompiledQueueItems, locationIndexByDispId


def _calculateGridKeys(preparedDisps: List[PreparedDisp], coordSetById, textStyleById):
    """ Calculate Grid Keys

    This method Determines which grids this disp will live in,
    and creates GridKeyIndex entries for those grid keys for this disp.

    """

    startTime = datetime.now(pytz.utc)

    # ---------------
    # inserts for GridKeyCompilerQueue
    gridCompiledQueueItems = set()

    # GridKeyIndexes to insert
    gridKeyIndexesByDispId = defaultdict(list)

    # Create a small piece of reusable code
    def addGridKey(disp, gridKey):
        # Create the compiler queue item
        gridCompiledQueueItems.add(
            CoordSetIdGridKeyData(coordSetId=disp.coordSetId,
                                  gridKey=gridKey)
        )

        # Create the grid key index item
        gridKeyIndexesByDispId[disp.id].append(
            dict(dispId=disp.id,
                 coordSetId=disp.coordSetId,
                 gridKey=gridKey,
                 importGroupHash=disp.importGroupHash))

    for pdisp in preparedDisps:
        # Get a reference to the coordSet
        coordSet = coordSetById[pdisp.disp.coordSetId]

        if isinstance(pdisp.disp, DispGroup):
            if pdisp.disp.compileAsTemplate:
                addGridKey(pdisp.disp, makeDispGroupGridKey(coordSet.id))
            continue

        # Calculate the grid keys
        gridKeys = makeGridKeysForDisp(coordSet, pdisp.disp, pdisp.geomArray,
                                       textStyleById)

        for gridKey in gridKeys:
            addGridKey(pdisp.disp, gridKey)

    logger.debug("Calculated GridKeys for %s disps in %s",
                 len(preparedDisps), (datetime.now(pytz.utc) - startTime))

    return gridCompiledQueueItems, gridKeyIndexesByDispId


def _updateDispsJson(preparedDisps: List[PreparedDisp]):
    """ Update Disps Json

    This assigns the updated dispJson from the dispDict

    """
    for pdisp in preparedDisps:
        stripped = _packDispJson(pdisp.disp, pdisp.dispDict)

        # Dump the json back into the disp
        pdisp.disp.dispJson = json.dumps(stripped)


def _packDispJson(disp, dispDict) -> Dict:
    """ Pack Disp Json

    This method creates the disp hash id.

    It assigns the hashId to the disp and dispDict

    :param disp: The disp to assign the data back to
    :param dispDict: The updated dispDict prepared elsewhere
        The only updated done outside of this method is for the DispGroup
    :return: The updated dispDict (This result is used by the BranchDispUpdater)

    """
    # Strip out the nulls and falses, to make it even more compact
    stripped = {k: v
                for k, v in dispDict.items()
                if v is not None and v is not False}

    hashId = _createHashId(stripped)

    # Assign the value
    stripped['hid'] = hashId

    # Write the "compiled" disp JSON back to the disp.
    disp.hashId = hashId

    return stripped


def _insertToDb(dispIds, gridCompiledQueueItems, gridKeyIndexesByDispId,
                locationCompiledQueueItems, locationIndexByDispId, queueIds):
    """ Insert to DB

    This method provides the DB inserts and deletes after the data has been calculated.

    """
    startTime = datetime.now(pytz.utc)

    dispBaseTable = DispBase.__table__
    dispQueueTable = DispIndexerQueue.__table__

    gridKeyIndexTable = GridKeyIndex.__table__
    gridQueueTable = GridKeyCompilerQueue.__table__

    locationIndexTable = LocationIndex.__table__
    locationIndexCompilerQueueTable = LocationIndexCompilerQueue.__table__

    engine = CeleryDbConn.getDbEngine()
    conn = engine.connect()
    transaction = conn.begin()
    try:
        lockedDispIds = conn.execute(Select(
            whereclause=dispBaseTable.c.id.in_(dispIds),
            columns=[dispBaseTable.c.id],
            for_update=True))

        lockedDispIds = [o[0] for o in lockedDispIds]

        # Ensure that the Disps exist, otherwise we get an integrity error.
        gridKeyIndexes = []
        locationIndexes = []
        for dispId in lockedDispIds:
            gridKeyIndexes.extend(gridKeyIndexesByDispId[dispId])

            if dispId in locationIndexByDispId:
                locationIndexes.append(locationIndexByDispId[dispId])

        # Delete existing items in the location and grid index

        # grid index
        conn.execute(
            gridKeyIndexTable.delete(gridKeyIndexTable.c.dispId.in_(dispIds))
        )

        # location index
        conn.execute(
            locationIndexTable.delete(locationIndexTable.c.dispId.in_(dispIds))
        )

        # ---------------
        # Insert the Grid Key indexes
        if gridKeyIndexes:
            conn.execute(gridKeyIndexTable.insert(), gridKeyIndexes)

        # Directly insert into the Grid compiler queue.
        if gridCompiledQueueItems:
            conn.execute(
                gridQueueTable.insert(),
                [dict(coordSetId=i.coordSetId, gridKey=i.gridKey)
                 for i in gridCompiledQueueItems]
            )

        # ---------------
        # Insert the Location indexes
        if locationIndexes:
            conn.execute(locationIndexTable.insert(), locationIndexes)

        # Directly insert into the Location compiler queue.
        if locationCompiledQueueItems:
            conn.execute(
                locationIndexCompilerQueueTable.insert(),
                [dict(modelSetId=i.modelSetId, indexBucket=i.indexBucket)
                 for i in locationCompiledQueueItems]
            )

        # ---------------
        # Finally, delete the disp queue items

        conn.execute(
            dispQueueTable.delete(dispQueueTable.c.id.in_(queueIds))
        )

        transaction.commit()
        logger.debug("Committed %s GridKeyIndex in %s",
                     len(gridKeyIndexes), (datetime.now(pytz.utc) - startTime))

    except Exception as e:
        raise

    finally:
        conn.close()



from datetime import datetime

from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from vortex.Tuple import addTupleType, TupleField, Tuple


@addTupleType
class LocationIndexTuple(Tuple):
    __tupleType__ = diagramTuplePrefix + "LocationIndexTuple"

    modelSetKey: str = TupleField()
    indexBucket: str = TupleField()

    # The compressed (deflated) json string.
    jsonStr: str = TupleField()
    lastUpdate: str = TupleField()



""" 
 * orm.Display.py
 *
 *  Copyright Synerty Pty Ltd 2011
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
"""
import typing

from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from sqlalchemy import Column, orm
from sqlalchemy import ForeignKey
from sqlalchemy import Integer, String, Boolean
from sqlalchemy.orm import relationship
from sqlalchemy.orm.mapper import reconstructor
from sqlalchemy.sql.schema import Index, Sequence
from sqlalchemy.sql.sqltypes import Float, DateTime
from vortex.Tuple import Tuple, addTupleType, TupleField, JSON_EXCLUDE

from .DeclarativeBase import DeclarativeBase
from .ModelSet import ModelCoordSet, ModelSet

DISP_SHORT_ATTR_NAME_MAP = {'colorId': 'c',
                            'fillColorId': 'fc',
                            'lineColorId': 'lc',
                            'lineStyleId': 'ls',
                            'lineWidth': 'w',
                            'text': 'te',
                            'groupId': 'gi',
                            'targetGroupId': 'tg',
                            }


@addTupleType
class DispLayer(Tuple, DeclarativeBase):
    __tablename__ = 'DispLayer'
    __tupleTypeShort__ = 'DLA'
    __tupleType__ = diagramTuplePrefix + __tablename__

    #: Misc data holder
    data = TupleField()

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(50), nullable=False)
    order = Column(Integer, nullable=False, server_default='0')
    selectable = Column(Boolean, nullable=False, server_default='false')
    visible = Column(Boolean, nullable=False, server_default='true')

    modelSetId = Column(Integer, ForeignKey('ModelSet.id', ondelete='CASCADE'),
                        nullable=False)
    modelSet = relationship(ModelSet)

    importHash = Column(String(100), doc=JSON_EXCLUDE)

    __table_args__: typing.Tuple = (
        Index("idx_DispLayer_modelSetId", modelSetId, unique=False),
        Index("idx_DispLayer_importHash", modelSetId, importHash, unique=True),
    )


@addTupleType
class DispLevel(Tuple, DeclarativeBase):
    __tablename__ = 'DispLevel'
    __tupleTypeShort__ = 'DLE'
    __tupleType__ = diagramTuplePrefix + __tablename__

    #: Misc data holder
    data = TupleField()

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(50), nullable=False)
    order = Column(Integer, nullable=False, server_default='0')
    minZoom = Column(Float)
    maxZoom = Column(Float)

    coordSetId = Column(Integer,
                        ForeignKey('ModelCoordSet.id', ondelete='CASCADE'),
                        nullable=False)
    coordSet = relationship(ModelCoordSet, foreign_keys=[coordSetId])

    importHash = Column(String(100), doc=JSON_EXCLUDE)

    __table_args__ = (
        Index("idx_DispLevel_coordSetId", coordSetId, unique=False),
        Index("idx_DispLevel_importHash", coordSetId, importHash, unique=True),
    )


@addTupleType
class DispTextStyle(Tuple, DeclarativeBase):
    __tupleTypeShort__ = 'DTS'
    __tablename__ = 'DispTextStyle'
    __tupleType__ = diagramTuplePrefix + __tablename__

    #: Misc data holder
    data = TupleField()

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(50), nullable=False)
    fontName = Column(String(30), nullable=False, server_default="GillSans")
    fontSize = Column(Integer, nullable=False, server_default='9')
    fontStyle = Column(String(30))
    scalable = Column(Boolean, nullable=False, server_default="true")
    scaleFactor = Column(Integer, nullable=False, server_default="1")

    modelSetId = Column(Integer, ForeignKey('ModelSet.id', ondelete='CASCADE'),
                        doc=JSON_EXCLUDE, nullable=False)
    modelSet = relationship(ModelSet)

    importHash = Column(String(100), doc=JSON_EXCLUDE)

    __table_args__ = (
        Index("idx_DispTextStyle_modelSetId", modelSetId, unique=False),
        Index("idx_DispTextStyle_importHash", modelSetId, importHash, unique=True),
    )


@addTupleType
class DispLineStyle(Tuple, DeclarativeBase):
    __tupleTypeShort__ = 'DLS'
    __tablename__ = 'DispLineStyle'
    __tupleType__ = diagramTuplePrefix + __tablename__

    #: Misc data holder
    data = TupleField()

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(50), nullable=False)
    backgroundFillDashSpace = Column(Boolean, nullable=False, server_default='false')
    capStyle = Column(String(15), nullable=False)
    joinStyle = Column(String(15), nullable=False)
    dashPattern = Column(String(50))
    startArrowSize = Column(Integer)
    endArrowSize = Column(Integer)
    winStyle = Column(Integer, nullable=False)

    modelSetId = Column(Integer, ForeignKey('ModelSet.id', ondelete='CASCADE'),
                        doc=JSON_EXCLUDE, nullable=False)
    modelSet = relationship(ModelSet)

    importHash = Column(String(100), doc=JSON_EXCLUDE)

    __table_args__ = (
        Index("idx_DispLineStyle_modelSetId", modelSetId, unique=False),
        Index("idx_DispLineStyle_importHash", modelSetId, importHash, unique=True),
    )


@addTupleType
class DispColor(Tuple, DeclarativeBase):
    __tupleTypeShort__ = 'DC'
    __tablename__ = 'DispColor'
    __tupleType__ = diagramTuplePrefix + __tablename__

    #: Misc data holder
    data = TupleField()

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(50), doc=JSON_EXCLUDE, nullable=False)
    color = Column(String(20), server_default='orange')
    altColor = Column(String(20))
    swapPeriod = Column(Float)

    modelSetId = Column(Integer, ForeignKey('ModelSet.id', ondelete='CASCADE'),
                        doc=JSON_EXCLUDE, nullable=False)
    modelSet = relationship(ModelSet)

    importHash = Column(String(100), doc=JSON_EXCLUDE)

    __table_args__ = (
        Index("idx_DispColor_modelSetId", modelSetId, unique=False),
        Index("idx_DispColor_importHash", modelSetId, importHash, unique=True),
    )


class DispBase(Tuple, DeclarativeBase):
    __tablename__ = 'DispBase'

    # Types
    # Must align with constants in javascript DispBase class
    GROUP = 10
    GROUP_PTR = 11
    TEXT = 40
    POLYGON = 50
    POLYLINE = 51
    ELLIPSE = 60
    NULL = 70

    id_seq = Sequence('DispBase_id_seq',
                      metadata=DeclarativeBase.metadata,
                      schema=DeclarativeBase.metadata.schema)
    id = Column(Integer, id_seq, server_default=id_seq.next_value(),
                primary_key=True, autoincrement=False)

    type = Column(Integer, doc=JSON_EXCLUDE, nullable=False)

    # ===== START BRANCH

    #: The branch that this Disp belongs to.
    branchId = Column(Integer, ForeignKey('BranchIndex.id', ondelete='CASCADE'),
                      doc='bi')

    #: The stage of the branch that this DISP
    branchStage = Column(Integer, doc='bs')

    #: This is the unique hash of the contents of this disp within this coordSetId.
    hashId = Column(String, doc='hid')

    #: The coordSetId+hashId that this disp replaces (for branches)
    # ForeignKey('DispBase.id', ondelete='SET NULL'),
    replacesHashId = Column(String, doc='rid')

    # ===== END BRANCH

    #: Used for disps that belong to a DispGroup
    groupId = Column(Integer, ForeignKey('DispBase.id', ondelete='CASCADE'),
                     doc='gi')

    coordSetId = Column(Integer, ForeignKey('ModelCoordSet.id', ondelete='CASCADE'),
                        doc=JSON_EXCLUDE,
                        nullable=False)
    coordSet = relationship(ModelCoordSet)

    #: Layer
    layerId = Column(Integer, ForeignKey('DispLayer.id'), doc='la')
    layer = relationship(DispLayer)

    #: Layer
    levelId = Column(Integer, ForeignKey('DispLevel.id'), doc='le')
    level = relationship(DispLevel)

    #: Order
    zOrder = Column(Integer, server_default='0', nullable=False, doc='z')

    # MAX_STR
    dispJson = Column(String, doc=JSON_EXCLUDE)

    #: The location of this Disp EG - [coordSetId, dispId, 0.0, 0.0] = [..., x, y]
    locationJson = Column(String, doc=JSON_EXCLUDE)

    #: Key, This value is a unique ID of the object that this graphic represents
    key = Column(String, doc='k')

    #: Selectable, Is is this item selectable?, the layer also needs selectable=true
    selectable = Column(Boolean, doc='s', nullable=False, server_default='false')

    #: Data, Generic data that is passed in the context for the item select popup
    dataJson = Column(String, doc='d')

    # THIS FIELD IS NOT USED ANYWHERE!!!!
    importUpdateDate = Column(DateTime(True), doc=JSON_EXCLUDE)

    importHash = Column(String, doc=JSON_EXCLUDE)
    importGroupHash = Column(String, doc=JSON_EXCLUDE)

    liveDbLinks = relationship("LiveDbDispLink")

    __mapper_args__ = {'polymorphic_on': type,
                       'with_polymorphic': '*'}

    __table_args__: typing.Tuple = (
        Index("idx_Disp_importGroupHash", importGroupHash, unique=False),
        # Index("idx_Disp_importHash", importHash, unique=True),
        Index("idx_Disp_importUpdateDate", importUpdateDate, unique=False),

        Index("idx_Disp_layerId", layerId, unique=False),
        Index("idx_Disp_levelId", levelId, unique=False),
        Index("idx_Disp_coordSetId_", coordSetId, unique=False),
        Index("idx_Disp_groupId", groupId, unique=False),
        Index("idx_Disp_branchId", branchId, unique=False),
        # Index("idx_Disp_hashId", coordSetId, hashId, unique=True),
        # Index("idx_Disp_replacesHashId", replacesHashId),
    )


@addTupleType
class DispNull(DispBase):
    """ Disp Null

    This display type does not display at all, It's an object that marks the deletion
    of a prior Disp by a branch.

    Disps in branches replace existing disps, thats how the store the change.
    This disp replaces an existing disp with nothing, with the effect of deleting it.

    This disp does have a geometry, which is equivilent to the bounding box of the shape
    it replaces.

    This whole approach allows the branch delete mechanism to slot into the existing
    diagram structures.

    """
    __tablename__ = 'DispNull'
    __tupleTypeShort__ = 'DN'
    __tupleType__ = diagramTuplePrefix + __tablename__

    RENDERABLE_TYPE = DispBase.NULL

    __mapper_args__ = {'polymorphic_identity': RENDERABLE_TYPE}

    id = Column(Integer, ForeignKey('DispBase.id', ondelete='CASCADE')
                , primary_key=True, autoincrement=False)

    geomJson = Column(String, nullable=False, doc='g')

    __table_args__: typing.Tuple = (
    )

    # noinspection PyMissingConstructor
    @orm.reconstructor
    def __init__(self):
        pass


@addTupleType
class DispText(DispBase):
    __tablename__ = 'DispText'
    __tupleTypeShort__ = 'DT'
    __tupleType__ = diagramTuplePrefix + __tablename__

    RENDERABLE_TYPE = DispBase.TEXT

    __mapper_args__ = {'polymorphic_identity': RENDERABLE_TYPE}

    id = Column(Integer, ForeignKey('DispBase.id', ondelete='CASCADE')
                , primary_key=True, autoincrement=False)

    verticalAlign = Column(Integer, doc='va', nullable=False, server_default='-1')
    horizontalAlign = Column(Integer, doc='ha', nullable=False, server_default='0')
    rotation = Column(Float, doc='r', nullable=False, server_default='0')
    text = Column(String, doc='te', nullable=True, server_default="new text label")
    textFormat = Column(String(1000), doc=JSON_EXCLUDE, nullable=True)

    textHeight = Column(Float, doc='th', nullable=True)
    textHStretch = Column(Float, doc='hs', nullable=False, server_default="1")

    geomJson = Column(String, nullable=False, doc='g')

    colorId = Column(Integer, ForeignKey('DispColor.id'), doc='c')
    color = relationship(DispColor)

    textStyleId = Column(Integer, ForeignKey('DispTextStyle.id'),
                         doc='fs', nullable=False)
    textStyle = relationship(DispTextStyle)

    __table_args__: typing.Tuple = (
        Index("idx_DispText_colorId", colorId, unique=False),
        Index("idx_DispText_styleId", textStyleId, unique=False),
    )

    # noinspection PyMissingConstructor
    @orm.reconstructor
    def __init__(self):
        pass


@addTupleType
class DispPolygon(DispBase):
    __tablename__ = 'DispPolygon'
    __tupleTypeShort__ = 'DPG'
    __tupleType__ = diagramTuplePrefix + __tablename__

    RENDERABLE_TYPE = DispBase.POLYGON
    __mapper_args__ = {'polymorphic_identity': RENDERABLE_TYPE}

    id = Column(Integer, ForeignKey('DispBase.id', ondelete='CASCADE')
                , primary_key=True, autoincrement=False)

    cornerRadius = Column(Float, doc='cr', nullable=False, server_default='0')
    lineWidth = Column(Integer, doc='w', nullable=False, server_default='2')

    geomJson = Column(String, nullable=False, doc='g')

    fillColorId = Column(Integer, ForeignKey('DispColor.id'), doc='fc')
    fillColor = relationship(DispColor, foreign_keys=fillColorId)

    FILL_TOP_TO_BOTTOM = 0
    FILL_BOTTOM_TO_TOP = 1
    FILL_RIGHT_TO_LEFT = 2
    FILL_LEFT_TO_RIGHT = 3
    fillDirection = Column(Integer, doc='fd')
    fillPercent = Column(Float, doc='fp')
    isRectangle = Column(Boolean, doc='r')

    lineColorId = Column(Integer, ForeignKey('DispColor.id'), doc='lc')
    lineColor = relationship(DispColor, foreign_keys=lineColorId)

    lineStyleId = Column(Integer, ForeignKey('DispLineStyle.id'),
                         doc='ls')
    lineStyle = relationship(DispLineStyle)

    __table_args__ = (
        Index("idx_DispPolygon_fillColorId", fillColorId, unique=False),
        Index("idx_DispPolygon_lineColorId", lineColorId, unique=False),
        Index("idx_DispPolygon_lineStyleId", lineStyleId, unique=False),
    )

    # noinspection PyMissingConstructor
    @orm.reconstructor
    def __init__(self):
        pass


@addTupleType
class DispPolyline(DispBase):
    __tablename__ = 'DispPolyline'
    __tupleTypeShort__ = 'DPL'
    __tupleType__ = diagramTuplePrefix + __tablename__

    RENDERABLE_TYPE = DispBase.POLYLINE
    __mapper_args__ = {'polymorphic_identity': RENDERABLE_TYPE}

    id = Column(Integer, ForeignKey('DispBase.id', ondelete='CASCADE')
                , primary_key=True, autoincrement=False)

    lineWidth = Column(Integer, doc='w', nullable=False, server_default='2')

    geomJson = Column(String, nullable=False, doc='g')

    lineColorId = Column(Integer, ForeignKey('DispColor.id'), doc='lc')
    lineColor = relationship(DispColor, foreign_keys=lineColorId)

    lineStyleId = Column(Integer, ForeignKey('DispLineStyle.id'), doc='ls')
    lineStyle = relationship(DispLineStyle)

    #: Start Key, The key of another disp object,
    # If the start point of this graphic is linked another disp obj
    startKey = Column(String(50), doc='sk')

    #: End Key, The key of another disp object,
    # If the end point of this graphic is linked another disp obj
    endKey = Column(String(50), doc='ek')

    #: Start end type, is this an arrow, etc?
    startEndType = Column(Integer, doc='st')

    #: End End Type, See Start end type
    endEndType = Column(Integer, doc='et')

    __table_args__ = (
        Index("idx_DispPolyline_lineColorId", lineColorId, unique=False),
        Index("idx_DispPolyline_lineStyleId", lineStyleId, unique=False),
        Index("idx_DispPolyline_startKey", startKey, unique=False),
        Index("idx_DispPolyline_endKey", endKey, unique=False),
    )

    # noinspection PyMissingConstructor
    @orm.reconstructor
    def __init__(self):
        pass


@addTupleType
class DispEllipse(DispBase):
    __tablename__ = 'DispEllipse'
    __tupleTypeShort__ = 'DE'
    __tupleType__ = diagramTuplePrefix + __tablename__

    RENDERABLE_TYPE = DispBase.ELLIPSE
    __mapper_args__ = {'polymorphic_identity': RENDERABLE_TYPE}

    id = Column(Integer, ForeignKey('DispBase.id', ondelete='CASCADE')
                , primary_key=True, autoincrement=False)

    xRadius = Column(Float, doc='xr', nullable=False, server_default='10.0')
    yRadius = Column(Float, doc='yr', nullable=False, server_default='10.0')
    rotation = Column(Float, doc='r', nullable=False, server_default='0')
    startAngle = Column(Float, doc='sa', nullable=False, server_default='0')
    endAngle = Column(Float, doc='ea', nullable=False, server_default='360')
    lineWidth = Column(Integer, doc='w', nullable=False, server_default='2')

    geomJson = Column(String, nullable=False, doc='g')

    fillColorId = Column(Integer, ForeignKey('DispColor.id'), doc='fc')
    fillColor = relationship(DispColor, foreign_keys=fillColorId)

    lineColorId = Column(Integer, ForeignKey('DispColor.id'), doc='lc')
    lineColor = relationship(DispColor, foreign_keys=lineColorId)

    lineStyleId = Column(Integer, ForeignKey('DispLineStyle.id'),
                         doc='ls')
    lineStyle = relationship(DispLineStyle)

    __table_args__ = (
        Index("idx_DispEllipse_fillColorId", fillColorId, unique=False),
        Index("idx_DispEllipse_lineColorId", lineColorId, unique=False),
        Index("idx_DispEllipse_lineStyleId", lineStyleId, unique=False),
    )

    # noinspection PyMissingConstructor
    @orm.reconstructor
    def __init__(self):
        pass


@addTupleType
class DispGroup(DispBase):
    __tablename__ = 'DispGroup'
    __tupleTypeShort__ = 'DG'
    __tupleType__ = diagramTuplePrefix + __tablename__

    RENDERABLE_TYPE = DispBase.GROUP
    __mapper_args__ = {'polymorphic_identity': RENDERABLE_TYPE}

    id = Column(Integer, ForeignKey('DispBase.id', ondelete='CASCADE')
                , primary_key=True, autoincrement=False)

    # TODO: this needs to be unique WITH the coordSetId
    name = Column(String, doc='n', nullable=False)
    compileAsTemplate = Column(Boolean, doc=JSON_EXCLUDE,
                               nullable=False, server_default='false')

    disps = relationship(DispBase,
                         primaryjoin='DispBase.groupId==DispGroup.id',
                         foreign_keys=[DispBase.groupId],
                         remote_side='DispBase.groupId')

    #: This field stores an array of disps that belong to this group.
    dispsForJson = TupleField(shortName='di')

    @reconstructor
    def __init__(self):
        DispBase.__init__(self)


@addTupleType
class DispGroupPointer(DispBase):
    __tablename__ = 'DispGroupPointer'
    __tupleTypeShort__ = 'DGP'
    __tupleType__ = diagramTuplePrefix + __tablename__

    RENDERABLE_TYPE = DispBase.GROUP_PTR
    __mapper_args__ = {'polymorphic_identity': RENDERABLE_TYPE}

    id = Column(Integer, ForeignKey('DispBase.id', ondelete='CASCADE')
                , primary_key=True, autoincrement=False)

    rotation = Column(Integer, doc='r', server_default='0', nullable=False)

    verticalScale = Column(Float, doc='vs', nullable=False, server_default='1.0')
    horizontalScale = Column(Float, doc='hs', nullable=False, server_default='1.0')

    geomJson = Column(String, nullable=False, doc='g')

    targetDispGroupId = Column(Integer, ForeignKey('DispGroup.id', ondelete='SET NULL'),
                               doc='tg')

    targetDispGroupName = Column(String, doc='tn')

    disps = relationship(DispBase,
                         primaryjoin='DispBase.groupId==DispGroupPointer.id',
                         foreign_keys=[DispBase.groupId],
                         remote_side='DispBase.groupId')

    __table_args__ = (
        Index("idxDispGroupPointer_targetDispGroupId", targetDispGroupId, unique=False),
    )

    # noinspection PyMissingConstructor
    @orm.reconstructor
    def __init__(self):
        pass



from geoalchemy2 import WKBElement
from typing import Optional, List

from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from peek_plugin_diagram.tuples.model.ImportLiveDbDispLinkTuple import \
    ImportLiveDbDispLinkTuple
from vortex.Tuple import Tuple, addTupleType, TupleField


@addTupleType
class ImportDispGroupTuple(Tuple):
    """ Imported Display Group

    This tuple is used by other plugins to load groups of display objects
    into the diagram model.

    Display items are apart of this group if their :code:`parentDispGroupHash` value is
    set to the :code:`importHash` of this group.

    The display items that are apart of this group will never be displayed.

    To create an instance of this display group on the diagram,
    import a ImportDispGroupPtyTuple display object.

    """
    __tupleType__ = diagramTuplePrefix + 'ImportDispGroupTuple'

    ### BEGIN DISP COMMON FIELDS ###

    #: Selectable, Is is this item selectable?, the layer also needs selectable=true
    selectable: bool = TupleField()

    #: Data, Generic data, this is passed to the popup context in the UI.
    # peek_plugin_diagram doesn't care as long as it's json compatible or None
    # Json length Length = 400
    data: Optional[dict] = TupleField(None)

    #: The unique hash of this display object
    # This will be referenced by:
    # ImportDispGroupPrtTuple.targetDispGroupHash, and
    # Any disps in the group, disp.parentDispGroupHash
    importHash: str = TupleField()

    #: The unique hash for all the display items imported in a group with this one.
    #: for example, a page or tile reference.
    importGroupHash: str = TupleField()

    #: The key of the ModelSet to import into
    modelSetKey: str = TupleField()

    #: The Key of the Coordinate Set to import into
    coordSetKey: str = TupleField()

    ### BEGIN FIELDS FOR THIS DISP ###

    #: A name for this dispGroup
    name: str = TupleField()

    # If this is set to true, then this group will be available in the user interface
    # for edit support (located in the special '##|dispgroup' grid key)
    compileAsTemplate: bool = TupleField(False)




import logging
from datetime import datetime
from typing import List, Dict, Tuple

import pytz
import ujson as json
from peek_plugin_base.storage.DbConnection import pgCopyInsert, convertToCoreSqlaInsert
from peek_plugin_base.worker import CeleryDbConn
from peek_plugin_diagram._private.server.controller.DispCompilerQueueController import \
    DispCompilerQueueController
from peek_plugin_diagram._private.storage.Display import \
    DispBase, DispEllipse, DispPolygon, DispText, DispPolyline, DispGroup, \
    DispGroupPointer, DispNull
from peek_plugin_diagram._private.storage.ModelSet import \
    ModelCoordSet, getOrCreateCoordSet
from peek_plugin_diagram._private.worker.CeleryApp import celeryApp
from peek_plugin_diagram._private.worker.tasks.ImportDispLink import importDispLinks
from peek_plugin_diagram._private.worker.tasks.LookupHashConverter import \
    LookupHashConverter
from peek_plugin_diagram.tuples.shapes.ImportDispEllipseTuple import \
    ImportDispEllipseTuple
from peek_plugin_diagram.tuples.shapes.ImportDispGroupPtrTuple import \
    ImportDispGroupPtrTuple
from peek_plugin_diagram.tuples.shapes.ImportDispGroupTuple import ImportDispGroupTuple
from peek_plugin_diagram.tuples.shapes.ImportDispPolygonTuple import \
    ImportDispPolygonTuple
from peek_plugin_diagram.tuples.shapes.ImportDispPolylineTuple import \
    ImportDispPolylineTuple
from peek_plugin_diagram.tuples.shapes.ImportDispTextTuple import ImportDispTextTuple
from peek_plugin_livedb.tuples.ImportLiveDbItemTuple import ImportLiveDbItemTuple
from txcelery.defer import DeferrableTask
from vortex.Payload import Payload

logger = logging.getLogger(__name__)

IMPORT_TUPLE_MAP = {
    ImportDispGroupTuple.tupleType(): DispGroup,
    ImportDispGroupPtrTuple.tupleType(): DispGroupPointer,
    ImportDispEllipseTuple.tupleType(): DispEllipse,
    ImportDispPolygonTuple.tupleType(): DispPolygon,
    ImportDispPolylineTuple.tupleType(): DispPolyline,
    ImportDispTextTuple.tupleType(): DispText
}

IMPORT_FIELD_NAME_MAP = {
    'levelHash': 'levelId',
    'layerHash': 'layerId',
    'lineStyleHash': 'lineStyleId',
    'colorHash': 'colorId',
    'fillColorHash': 'fillColorId',
    'lineColorHash': 'lineColorId',
    'textStyleHash': 'textStyleId',
    'targetDispGroupHash': 'targetDispGroupId',
    'targetDispGroupName': 'targetDispGroupName'
}

IMPORT_SORT_ORDER = {
    ImportDispGroupTuple.tupleType(): 0,
    ImportDispGroupPtrTuple.tupleType(): 1,
    ImportDispEllipseTuple.tupleType(): 2,
    ImportDispPolygonTuple.tupleType(): 2,
    ImportDispPolylineTuple.tupleType(): 2,
    ImportDispTextTuple.tupleType(): 2
}


@DeferrableTask
@celeryApp.task(bind=True)
def importDispsTask(self, modelSetKey: str, coordSetKey: str,
                    importGroupHash: str,
                    dispsEncodedPayload: bytes) -> List[ImportLiveDbItemTuple]:
    """ Import Disp Task

    :returns None

    """
    try:
        importDisps = Payload().fromEncodedPayload(dispsEncodedPayload).tuples

        _validateImportDisps(importDisps)

        coordSet = _loadCoordSet(modelSetKey, coordSetKey)

        _validateImportDisps(importDisps)

        dispIdsToCompile, dispLinkImportTuples, ormDisps = _importDisps(
            coordSet, importDisps
        )

        _validateConvertedDisps(ormDisps)

        # Update the coord set view start position if required.
        _updateCoordSetPosition(coordSet, importDisps)

        _bulkLoadDispsTask(importGroupHash, ormDisps)

        liveDbImportTuples = importDispLinks(
            coordSet, importGroupHash, dispLinkImportTuples
        )

        DispCompilerQueueController.queueDispIdsToCompile(
            dispIdsToCompile, CeleryDbConn.getDbSession
        )

        return liveDbImportTuples

    except Exception as e:
        logger.exception(e)
        logger.debug("Retrying import displays, %s", e)
        raise self.retry(exc=e, countdown=3)


def _loadCoordSet(modelSetKey, coordSetKey):
    ormSession = CeleryDbConn.getDbSession()
    try:
        coordSet = getOrCreateCoordSet(ormSession, modelSetKey, coordSetKey)
        ormSession.expunge_all()
        return coordSet

    finally:
        ormSession.close()


def _validateImportDisps(importDisp: List):

    for importDisp in importDisp:
        isGroup = isinstance(importDisp, ImportDispGroupTuple)
        # isGroupChild = not isGroup and importDisp.parentDispGroupHash

        if not isGroup and importDisp.layerHash is None:
            raise Exception("Disps must have layers and levels")

        if not isGroup and importDisp.levelHash is None:
            raise Exception("Disps must have layers and levels")


def _validateConvertedDisps(disps: List):
    NoneT = type(None)

    def checkInt(importDisp, attrName):
        if hasattr(importDisp, attrName):
            if type(getattr(importDisp, attrName)) not in (int, float, NoneT):
                raise Exception('Disp %s must be int : "%s"'
                                % (attrName, importDisp.colorHash))

    for disp in disps:
        checkInt(disp, 'colorHash')
        checkInt(disp, 'lineColorHash')
        checkInt(disp, 'fillColorHash')
        checkInt(disp, 'lineStyleHash')
        checkInt(disp, 'textStyleHash')
        checkInt(disp, 'layerHash')
        checkInt(disp, 'levelHash')


def _importDisps(coordSet: ModelCoordSet, importDisps: List):
    """ Link Disps

    1) Use the AgentImportDispGridLookup to convert lookups from importHash
        to id
    2) set the  coordSetId

    This is not done in a thread because the lookups cause issues

    """

    dispIdGen = CeleryDbConn.prefetchDeclarativeIds(DispBase, len(importDisps))

    dispIdsToCompile = []
    importDispLinks = []
    ormDisps = []

    ormSession = CeleryDbConn.getDbSession()
    try:

        lookupConverter = LookupHashConverter(ormSession,
                                              modelSetId=coordSet.modelSetId,
                                              coordSetId=coordSet.id)

        dispGroupPtrWithTargetHash: List[Tuple[DispGroupPointer, str]] = []
        dispGroupChildWithTargetHash: List[Tuple[DispBase, str]] = []

        # Preload any groups our pointers may point to.

        # Pre-import any DispGroup IDs we may need
        dispGroupTargetImportHashes = [
            o.targetDispGroupHash
            for o in importDisps
            if o.tupleType() == ImportDispGroupPtrTuple.tupleType()
        ]

        # This will store DispGroup and DispGroupPointer hashes
        groupIdByImportHash: Dict[str, int] = {
            o.importHash: o.id
            for o in
            ormSession.query(DispBase.importHash, DispBase.id)
                .filter(DispBase.importHash.in_(dispGroupTargetImportHashes))
                .filter(DispBase.coordSetId == coordSet.id)
        }

        del dispGroupTargetImportHashes

        # This is a list of DispGroup.id.
        # We use this to filter out disps that part of a DispGroup,
        # they don't get compiled
        dispGroupIds = set()

        # Sort the DispGroups first, so they are created before any FK references them
        sortedImportDisps = sorted(importDisps,
                                   key=lambda o: IMPORT_SORT_ORDER[o.tupleType()])

        for importDisp in sortedImportDisps:
            # Convert the geometry into the internal array format
            _convertGeom(importDisp)

            # Create the storage tuple instance, and copy over the data.
            ormDisp = _convertImportTuple(importDisp)
            ormDisps.append(ormDisp)

            # Preallocate the IDs for performance on PostGreSQL
            ormDisp.id = next(dispIdGen)

            # Assign the coord set id.
            ormDisp.coordSetId = coordSet.id

            # If this is a dispGroup, index it's ID
            if isinstance(ormDisp, DispGroup):
                dispGroupIds.add(ormDisp.id)
                groupIdByImportHash[ormDisp.importHash] = ormDisp.id

            # If this is a dispGroupPtr, index its targetHash so we can update it
            if isinstance(ormDisp, DispGroupPointer):
                groupIdByImportHash[ormDisp.importHash] = ormDisp.id

                if ormDisp.targetDispGroupName:
                    ormDisp.targetDispGroupName = '%s|%s' % (
                        coordSet.id, ormDisp.targetDispGroupName
                    )

                # Not all DispGroupPointers have targets,
                # they can be orphaned instances
                if importDisp.targetDispGroupHash:
                    dispGroupPtrWithTargetHash.append(
                        (ormDisp, importDisp.targetDispGroupHash)
                    )

            # If this is a dispGroupPtr, index its targetHash so we can update it
            parentDispGroupHash = getattr(importDisp, "parentDispGroupHash", None)
            if parentDispGroupHash:
                dispGroupChildWithTargetHash.append((ormDisp, parentDispGroupHash))

            # Add some interim data to the import display link, so it can be created
            if hasattr(importDisp, "liveDbDispLinks"):
                for importDispLink in importDisp.liveDbDispLinks:
                    attrName = importDispLink.dispAttrName
                    importDispLink.internalRawValue = getattr(ormDisp, attrName)
                    importDispLink.internalDispId = ormDisp.id
                    importDispLinks.append(importDispLink)

            # Convert the values of the liveDb attributes
            lookupConverter.convertLookups(ormDisp)

            # Add the after translate value, this is the Display Value
            if hasattr(importDisp, "liveDbDispLinks"):
                for importDispLink in importDisp.liveDbDispLinks:
                    attrName = importDispLink.dispAttrName
                    importDispLink.internalDisplayValue = getattr(ormDisp, attrName)

            # Queue the Disp to be compiled into a grid.
            # Disps belonging to a DispGroup do not get compiled into grids.
            if ormDisp.groupId not in dispGroupIds:
                dispIdsToCompile.append(ormDisp.id)

        # Link the DispGroups
        # Create the links between the Disp and DispGroup
        for ormDisp, groupHash in dispGroupChildWithTargetHash:
            groupOrmObjId = groupIdByImportHash.get(groupHash)
            if groupOrmObjId is None:
                raise Exception(
                    "DispGroup with importHash %s doesn't exist" % groupHash)

            ormDisp.groupId = groupOrmObjId

        # Link the DispGroupPtr to the DispGroup
        # This is only used when the dispGrouPtr points to a disp group
        for ormDisp, groupHash in dispGroupPtrWithTargetHash:
            groupOrmObjId = groupIdByImportHash.get(groupHash)
            if groupOrmObjId is None:
                raise Exception(
                    "DispGroup with importHash %s doesn't exist" % groupHash)

            ormDisp.targetDispGroupId = groupOrmObjId


    finally:
        ormSession.close()

    return dispIdsToCompile, importDispLinks, ormDisps


def _convertGeom(importDisp):
    if not hasattr(importDisp, "geom"):
        return

    coordArray = []
    for i in importDisp.geom:
        coordArray.append(i[0])
        coordArray.append(i[1])

    importDisp.geom = json.dumps(coordArray)


def _convertImportTuple(importDisp):
    """ Convert Import Tuple

    This method mostly copies over data from the import tuple into the storage tuple,
    converting some fields and field names as required.

    """
    if not importDisp.tupleType() in IMPORT_TUPLE_MAP:
        raise Exception(
            "Import Tuple %s is not a valid type" % importDisp.tupleType()
        )

    disp = IMPORT_TUPLE_MAP[importDisp.tupleType()]()

    for importFieldName in importDisp.tupleFieldNames():
        if importFieldName == "data" and importDisp.data:
            disp.dataJson = json.dumps(importDisp.data)
            continue

        if importFieldName == "geom":

            # Groups are stored in whichever grid where the where ever the center point
            # is located, OR, in the special dispgroup grid.
            if isinstance(disp, DispGroup) and not importDisp.geom:
                continue

            disp.geomJson = importDisp.geom

            # Moved to server, due to celery 3 pickle problem
            # disp.geomJson = json.dumps(convertFromWkbElement(importDisp.geom))
            continue

        # Convert the field name if it exists
        dispFieldName = IMPORT_FIELD_NAME_MAP.get(importFieldName, importFieldName)

        setattr(disp, dispFieldName, getattr(importDisp, importFieldName))

    if isinstance(disp, DispGroup):
        disp.dispsJson = '[]'

    return disp


def _bulkLoadDispsTask(importGroupHash: str, disps: List):
    """ Import Disps Links

    1) Drop all disps with matching importGroupHash

    2) set the  coordSetId

    :param importGroupHash:
    :param disps: An array of disp objects to import
    :return:
    """

    dispTable = DispBase.__table__

    engine = CeleryDbConn.getDbEngine()
    conn = engine.connect()
    transaction = conn.begin()

    try:
        conn.execute(dispTable
                     .delete()
                     .where(dispTable.c.importGroupHash == importGroupHash))

        transaction.commit()

        _bulkInsertDisps(engine, disps)

    except Exception:
        transaction.rollback()
        raise

    finally:
        conn.close()


def _bulkInsertDisps(engine, disps: List):
    """ Bulk Insert Disps

    1) Drop all disps with matching importGroupHash

    2) set the  coordSetId

    :param engine: The connection to use
    :param disps: An array of disp objects to import
    :return:
    """

    INSERT_MAP = (
        (DispGroup, (DispBase, DispGroup)),
        (DispGroupPointer, (DispBase, DispGroupPointer)),
        (DispEllipse, (DispBase, DispEllipse)),
        (DispPolyline, (DispBase, DispPolyline)),
        (DispPolygon, (DispBase, DispPolygon)),
        (DispText, (DispBase, DispText)),
        (DispNull, (DispBase, DispNull)),
    )

    startTime = datetime.now(pytz.utc)
    startDispsLen = len(disps)
    rawConn = engine.raw_connection()

    try:
        for DispType, Tables in INSERT_MAP:
            inserts = []
            remainingDisps = []

            for disp in disps:
                if isinstance(disp, DispType):
                    insertDict = convertToCoreSqlaInsert(disp, DispType)
                    insertDict['type'] = DispType.RENDERABLE_TYPE
                    inserts.append(insertDict)

                else:
                    remainingDisps.append(disp)

            disps = remainingDisps

            if not inserts:
                continue

            for Table in Tables:
                pgCopyInsert(rawConn, Table.__table__, inserts)
                # conn.execute(Table.__table__.insert(), inserts)

        if disps:
            raise Exception("_bulkInsertDisps: We didn't insert all the disps")

        logger.info("Inserted %s Disps in %s",
                    startDispsLen, (datetime.now(pytz.utc) - startTime))

        rawConn.commit()

    except Exception:
        rawConn.rollback()
        raise

    finally:
        rawConn.close()


def _updateCoordSetPosition(coordSet: ModelCoordSet, disps: List):
    """ Update CoordSet Position

    1) Drop all disps with matching importGroupHash

    2) set the  coordSetId

    :param coordSet:
    :param disps: An array of disp objects to import
    :return:
    """

    if coordSet.initialPanX or coordSet.initialPanY or coordSet.initialZoom:
        return

    startTime = datetime.now(pytz.utc)

    ormSession = CeleryDbConn.getDbSession()

    try:

        # Initialise the ModelCoordSet initial position if it's not set
        for disp in disps:
            if not hasattr(disp, 'geomJson'):
                continue
            coords = json.loads(disp.geomJson)
            coordSet.initialPanX = coords[0]
            coordSet.initialPanY = coords[1]
            coordSet.initialZoom = 0.05
            ormSession.merge(coordSet)
            break

        ormSession.commit()

        logger.info("Updated coordset position in %s",
                    (datetime.now(pytz.utc) - startTime))

    finally:
        ormSession.close()



import logging
from datetime import datetime
from typing import List

import pytz
from sqlalchemy import asc
from twisted.internet import task
from twisted.internet.defer import inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger, vortexLogFailure

from peek_plugin_base.storage.StorageUtil import makeCoreValuesSubqueryCondition
from peek_plugin_diagram._private.server.client_handlers.ClientGridUpdateHandler import \
    ClientGridUpdateHandler
from peek_plugin_diagram._private.server.controller.StatusController import \
    StatusController
from peek_plugin_diagram._private.storage.GridKeyIndex import \
    GridKeyCompilerQueue

logger = logging.getLogger(__name__)


class GridKeyCompilerQueueController:
    """ Grid Compiler

    Compile the disp items into the grid data

    1) Query for queue
    2) Process queue
    3) Delete from queue

    """

    FETCH_SIZE = 5
    PERIOD = 0.200

    QUEUE_MAX = 100
    QUEUE_MIN = 30

    def __init__(self, ormSessionCreator,
                 statusController: StatusController,
                 clientGridUpdateHandler: ClientGridUpdateHandler):
        self._ormSessionCreator = ormSessionCreator
        self._statusController: StatusController = statusController
        self._clientGridUpdateHandler: ClientGridUpdateHandler = clientGridUpdateHandler

        self._pollLoopingCall = task.LoopingCall(self._poll)
        self._lastQueueId = -1
        self._queueCount = 0

    def start(self):
        self._statusController.setGridCompilerStatus(True, self._queueCount)
        d = self._pollLoopingCall.start(self.PERIOD, now=False)
        d.addCallbacks(self._timerCallback, self._timerErrback)

    def isBusy(self) -> bool:
        return self._queueCount != 0

    def _timerErrback(self, failure):
        vortexLogFailure(failure, logger)
        self._statusController.setGridCompilerStatus(False, self._queueCount)
        self._statusController.setGridCompilerError(str(failure.value))

    def _timerCallback(self, _):
        self._statusController.setGridCompilerStatus(False, self._queueCount)

    def stop(self):
        if self._pollLoopingCall.running:
            self._pollLoopingCall.stop()

    def shutdown(self):
        self.stop()

    @inlineCallbacks
    def _poll(self):
        from peek_plugin_diagram._private.worker.tasks.GridCompilerTask import \
            compileGrids

        # We queue the grids in bursts, reducing the work we have to do.
        if self._queueCount > self.QUEUE_MIN:
            return

        # Check for queued items
        queueItems = yield self._grabQueueChunk()
        if not queueItems:
            return

        # De duplicated queued grid keys
        # This is the reason why we don't just queue all the celery tasks in one go.
        # If we keep them in the DB queue, we can remove the duplicates
        # and there are lots of them
        queueIdsToDelete = []

        gridKeySet = set()
        for i in queueItems:
            if i.gridKey in gridKeySet:
                queueIdsToDelete.append(i.id)
            else:
                gridKeySet.add(i.gridKey)

        if queueIdsToDelete:
            # Delete the duplicates and requery for our new list
            yield self._deleteDuplicateQueueItems(queueIdsToDelete)
            queueItems = yield self._grabQueueChunk()

        # Send the tasks to the peek worker
        for start in range(0, len(queueItems), self.FETCH_SIZE):

            items = queueItems[start: start + self.FETCH_SIZE]

            try:
                d = compileGrids.delay(items)
                d.addCallback(self._pollCallback, datetime.now(pytz.utc), len(items))
                d.addErrback(self._pollErrback, datetime.now(pytz.utc))

            except Exception as e:
                logger.exception(e)
                return

            # Set the watermark
            self._lastQueueId = items[-1].id

            self._queueCount += 1
            if self._queueCount >= self.QUEUE_MAX:
                break

    @deferToThreadWrapWithLogger(logger)
    def _grabQueueChunk(self):
        session = self._ormSessionCreator()
        try:
            qry = (session.query(GridKeyCompilerQueue)
                   .order_by(asc(GridKeyCompilerQueue.id))
                   .filter(GridKeyCompilerQueue.id > self._lastQueueId)
                   .yield_per(500)
                   # .limit(self.FETCH_SIZE)
                   )

            queueItems = qry.all()
            session.expunge_all()

            return queueItems

        finally:
            session.close()

    @deferToThreadWrapWithLogger(logger)
    def _deleteDuplicateQueueItems(self, itemIds):
        session = self._ormSessionCreator()
        table = GridKeyCompilerQueue.__table__
        try:
            SIZE = 1000
            for start in range(0, len(itemIds), SIZE):
                chunkIds = itemIds[start: start + SIZE]

                session.execute(table.delete(makeCoreValuesSubqueryCondition(
                    session.bind, table.c.id, chunkIds
                )))

            session.commit()
        finally:
            session.close()

    def _pollCallback(self, gridKeys: List[str], startTime, processedCount):
        self._queueCount -= 1
        logger.debug("Time Taken = %s" % (datetime.now(pytz.utc) - startTime))
        self._clientGridUpdateHandler.sendGrids(gridKeys)
        self._statusController.addToGridCompilerTotal(processedCount)
        self._statusController.setGridCompilerStatus(True, self._queueCount)

    def _pollErrback(self, failure, startTime):
        self._queueCount -= 1
        self._statusController.setGridCompilerError(str(failure.value))
        self._statusController.setGridCompilerStatus(True, self._queueCount)
        logger.debug("Time Taken = %s" % (datetime.now(pytz.utc) - startTime))
        vortexLogFailure(failure, logger)



"""Added CoordSet dispGroup enabled

Peek Plugin Database Migration Script

Revision ID: f4dce3e782ec
Revises: 81c1d2809046
Create Date: 2019-05-17 14:37:07.220197

"""

# revision identifiers, used by Alembic.
revision = 'f4dce3e782ec'
down_revision = '81c1d2809046'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.add_column('ModelCoordSet',
                  sa.Column('dispGroupTemplatesEnabled', sa.Boolean(),
                            server_default='false', nullable=True),
                  schema='pl_diagram')

    op.execute('UPDATE "pl_diagram"."ModelCoordSet" SET "dispGroupTemplatesEnabled" = false ')

    op.alter_column('ModelCoordSet', 'dispGroupTemplatesEnabled', type_=sa.Boolean(),
                    nullable=False, schema='pl_diagram')


def downgrade():
    op.drop_column('ModelCoordSet', 'dispGroupTemplatesEnabled', schema='pl_diagram')


import logging
from collections import defaultdict
from typing import List, Dict

from twisted.internet.defer import DeferredList, inlineCallbacks, Deferred

from peek_plugin_diagram._private.PluginNames import diagramFilt
from peek_plugin_diagram._private.client.controller.BranchIndexCacheController import \
    BranchIndexCacheController
from peek_plugin_diagram._private.tuples.branch.BranchIndexUpdateDateTuple import \
    BranchIndexUpdateDateTuple
from vortex.DeferUtil import vortexLogFailure
from vortex.Payload import Payload
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.VortexABC import SendVortexMsgResponseCallable
from vortex.VortexFactory import VortexFactory

logger = logging.getLogger(__name__)

clientBranchIndexWatchUpdateFromDeviceFilt = {
    'key': "clientBranchIndexWatchUpdateFromDevice"
}
clientBranchIndexWatchUpdateFromDeviceFilt.update(diagramFilt)


# ModelSet HANDLER
class BranchIndexCacheHandler(object):
    def __init__(self, cacheController: BranchIndexCacheController,
                 clientId: str):
        """ App BranchIndex Handler

        This class handles the custom needs of the desktop/mobile apps observing branchIndexs.

        """
        self._cacheController = cacheController
        self._clientId = clientId

        self._epObserve = PayloadEndpoint(
            clientBranchIndexWatchUpdateFromDeviceFilt, self._processObserve
        )

        self._uuidsObserving = set()

    def shutdown(self):
        self._epObserve.shutdown()
        self._epObserve = None

    # ---------------
    # Process update from the server

    def notifyOfBranchIndexUpdate(self, chunkKeys: List[str]):
        """ Notify of BranchIndex Updates

        This method is called by the client.BranchIndexCacheController when it receives updates
        from the server.

        """
        vortexUuids = set(VortexFactory.getRemoteVortexUuids()) & self._uuidsObserving
        self._uuidsObserving = vortexUuids

        payloadsByVortexUuid = defaultdict(Payload)

        for chunkKey in chunkKeys:
            encodedBranchIndexChunk = self._cacheController.branchIndexChunk(chunkKey)

            # Queue up the required client notifications
            for vortexUuid in vortexUuids:
                logger.debug("Sending unsolicited branchIndex %s to vortex %s",
                             chunkKey, vortexUuid)
                payloadsByVortexUuid[vortexUuid].tuples.append(encodedBranchIndexChunk)

        # Send the updates to the clients
        dl = []
        for vortexUuid, payload in list(payloadsByVortexUuid.items()):
            payload.filt = clientBranchIndexWatchUpdateFromDeviceFilt

            # Serialise in thread, and then send.
            d = payload.makePayloadEnvelopeDefer()
            d.addCallback(lambda payloadEnvelope: payloadEnvelope.toVortexMsgDefer())
            d.addCallback(VortexFactory.sendVortexMsg, destVortexUuid=vortexUuid)
            dl.append(d)

        # Log the errors, otherwise we don't care about them
        dl = DeferredList(dl, fireOnOneErrback=True)
        dl.addErrback(vortexLogFailure, logger, consumeError=True)

    # ---------------
    # Process observes from the devices
    @inlineCallbacks
    def _processObserve(self, payloadEnvelope: PayloadEnvelope,
                        vortexUuid: str,
                        sendResponse: SendVortexMsgResponseCallable,
                        **kwargs):

        payload = yield payloadEnvelope.decodePayloadDefer()

        updateDatesTuples: BranchIndexUpdateDateTuple = payload.tuples[0]

        self._uuidsObserving.add(vortexUuid)

        yield self._replyToObserve(payload.filt,
                                   updateDatesTuples.updateDateByChunkKey,
                                   sendResponse,
                                   vortexUuid)

    # ---------------
    # Reply to device observe

    @inlineCallbacks
    def _replyToObserve(self, filt,
                        lastUpdateByBranchIndexKey: Dict[str, str],
                        sendResponse: SendVortexMsgResponseCallable,
                        vortexUuid: str) -> None:
        """ Reply to Observe

        The client has told us that it's observing a new set of branchIndexs, and the lastUpdate
        it has for each of those branchIndexs. We will send them the branchIndexs that are out of date
        or missing.

        :param filt: The payload filter to respond to.
        :param lastUpdateByBranchIndexKey: The dict of branchIndexKey:lastUpdate
        :param sendResponse: The callable provided by the Vortex (handy)
        :returns: None

        """
        yield None

        branchIndexTuplesToSend = []

        # Check and send any updates
        for branchIndexKey, lastUpdate in lastUpdateByBranchIndexKey.items():
            if vortexUuid not in VortexFactory.getRemoteVortexUuids():
                logger.debug("Vortex %s is offline, stopping update")
                return

            # NOTE: lastUpdate can be null.
            encodedBranchIndexTuple = self._cacheController.branchIndexChunk(branchIndexKey)
            if not encodedBranchIndexTuple:
                logger.debug("BranchIndex %s is not in the cache" % branchIndexKey)
                continue

            # We are king, If it's it's not our version, it's the wrong version ;-)
            logger.debug("%s, %s,  %s",
                         encodedBranchIndexTuple.lastUpdate == lastUpdate,
                         encodedBranchIndexTuple.lastUpdate, lastUpdate)

            if encodedBranchIndexTuple.lastUpdate == lastUpdate:
                logger.debug("BranchIndex %s matches the cache" % branchIndexKey)
                continue

            branchIndexTuplesToSend.append(encodedBranchIndexTuple)
            logger.debug("Sending branchIndex %s from the cache" % branchIndexKey)

        # Send the payload to the frontend
        payload = Payload(filt=filt, tuples=branchIndexTuplesToSend)
        d: Deferred = payload.makePayloadEnvelopeDefer()
        d.addCallback(lambda payloadEnvelope: payloadEnvelope.toVortexMsgDefer())
        d.addCallback(sendResponse)
        d.addErrback(vortexLogFailure, logger, consumeError=True)



from vortex.Tuple import Tuple, addTupleType, TupleField

from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix


@addTupleType
class ImportDispLayerTuple(Tuple):
    """ Import Display Layer Tuple


    """
    __tupleType__ = diagramTuplePrefix + 'ImportDispLayerTuple'

    name: str = TupleField()

    order: int = TupleField()

    visible: bool = TupleField()

    selectable: bool = TupleField()

    importHash: str = TupleField()

    modelSetKey: str = TupleField()



import logging
from datetime import datetime
from typing import List

import pytz
from collections import namedtuple
from txcelery.defer import  DeferrableTask

from peek_plugin_base.worker import CeleryDbConn
from peek_plugin_diagram._private.storage.ModelSet import ModelSet
from peek_plugin_diagram._private.worker.CeleryApp import celeryApp
from peek_plugin_diagram._private.worker.tasks.LiveDbDisplayValueConverter import \
    LiveDbDisplayValueConverter
from peek_plugin_livedb.tuples.LiveDbDisplayValueUpdateTuple import \
    LiveDbDisplayValueUpdateTuple
from peek_plugin_livedb.tuples.LiveDbRawValueUpdateTuple import LiveDbRawValueUpdateTuple
from peek_plugin_livedb.worker.WorkerApi import WorkerApi

logger = logging.getLogger(__name__)

CoordSetIdGridKeyTuple = namedtuple("CoordSetIdGridKeyTuple", ["coordSetId", "gridKey"])
DispData = namedtuple('DispData', ['json', 'levelOrder', 'layerOrder'])


@DeferrableTask
@celeryApp.task
def convertLiveDbDisplayValuesTask(modelSetKey: str,
                                   rawUpdates: List[LiveDbRawValueUpdateTuple]):
    startTime = datetime.now(pytz.utc)
    keys = [u.key for u in rawUpdates]

    ormSession = CeleryDbConn.getDbSession()
    try:
        liveDbDataTypeLookup = WorkerApi.getLiveDbKeyDatatypeDict(
            ormSession, modelSetKey, keys
        )

        modelSetId = (ormSession.query(ModelSet.id)
                      .filter(ModelSet.key == modelSetKey)
                      .one()
                      .id)

        translater = LiveDbDisplayValueConverter.create(
            ormSession, modelSetId
        )

    finally:
        ormSession.close()

    displayValueUpdates: List[LiveDbDisplayValueUpdateTuple] = []

    for rawUpdate in rawUpdates:
        dataType = liveDbDataTypeLookup.get(rawUpdate.key)
        if dataType is None:
            logger.warning("LiveDB key %s is missing dataType", rawUpdate.key)
            continue

        displayValue = translater.translate(dataType, rawUpdate.rawValue)

        displayValueUpdates.append(LiveDbDisplayValueUpdateTuple(
            key=rawUpdate.key, displayValue=displayValue
        ))

    logger.debug("Converted %s LiveDB Raw Values in %s",
                 len(rawUpdates), (datetime.now(pytz.utc) - startTime))

    return displayValueUpdates



"""Added CoordSet GridSizes

Peek Plugin Database Migration Script

Revision ID: 6a97dd344aed
Revises: cbd5e370686f
Create Date: 2017-08-13 13:17:24.923154

"""

# revision identifiers, used by Alembic.
revision = '6a97dd344aed'
down_revision = 'cbd5e370686f'
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('ModelCoordSetGridSize',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('key', sa.Integer(), nullable=False),
                    sa.Column('min', sa.Float(), nullable=False),
                    sa.Column('max', sa.Float(), nullable=False),
                    sa.Column('xGrid', sa.Integer(), nullable=False),
                    sa.Column('yGrid', sa.Integer(), nullable=False),
                    sa.Column('coordSetId', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['coordSetId'],
                                            ['pl_diagram.ModelCoordSet.id'],
                                            ondelete='CASCADE'),
                    sa.PrimaryKeyConstraint('id'),
                    schema='pl_diagram'
                    )
    op.create_index('idx_CoordSetGridSize_key', 'ModelCoordSetGridSize',
                    ['coordSetId', 'key'], unique=True, schema='pl_diagram')

    op.add_column('ModelCoordSet',
                  sa.Column('smallestTextSize', sa.Float(), nullable=True, server_default='6.0'),
                  schema='pl_diagram')

    op.add_column('ModelCoordSet',
                  sa.Column('smallestShapeSize', sa.Float(), nullable=True, server_default='2.0'),
                  schema='pl_diagram')

    op.add_column('ModelCoordSet',
                  sa.Column('minZoom', sa.Float(), nullable=True, server_default='0.01'),
                  schema='pl_diagram')

    op.add_column('ModelCoordSet',
                  sa.Column('maxZoom', sa.Float(), nullable=True, server_default='10.0'),
                  schema='pl_diagram')

    op.execute('''
    UPDATE "pl_diagram"."ModelCoordSet"
    SET "smallestShapeSize" = 2.0,
        "smallestTextSize" = 6.0,
        "minZoom" = 0.01,
        "maxZoom" = 10.0
    ''')

    op.alter_column('ModelCoordSet', 'smallestTextSize',
                    existing_type=sa.Float(),
                    nullable=False,
                    schema='pl_diagram')

    op.alter_column('ModelCoordSet', 'smallestShapeSize',
                    existing_type=sa.Float(),
                    nullable=False,
                    schema='pl_diagram')

    op.alter_column('ModelCoordSet', 'minZoom',
                    existing_type=sa.Float(),
                    nullable=False,
                    schema='pl_diagram')

    op.alter_column('ModelCoordSet', 'maxZoom',
                    existing_type=sa.Float(),
                    nullable=False,
                    schema='pl_diagram')

    op.execute('''
        INSERT INTO "pl_diagram"."ModelCoordSetGridSize"
        ("coordSetId", "key", "min", "max", "xGrid", "yGrid")
        SELECT "id", 0 as "key", 0 as "min", 0.04 as "max", 30000 as "xGrid", 30000 as "yGrid"
        FROM "pl_diagram"."ModelCoordSet"
    ''')

    op.execute('''
        INSERT INTO "pl_diagram"."ModelCoordSetGridSize"
        ("coordSetId", "key", "min", "max", "xGrid", "yGrid")
        SELECT "id", 1 as "key", 0.04 as "min", 0.1 as "max", 10000 as "xGrid", 10000 as "yGrid"
        FROM "pl_diagram"."ModelCoordSet"
    ''')

    op.execute('''
        INSERT INTO "pl_diagram"."ModelCoordSetGridSize"
        ("coordSetId", "key", "min", "max", "xGrid", "yGrid")
        SELECT "id", 2 as "key", 0.1 as "min", 0.5 as "max", 2000 as "xGrid", 2000 as "yGrid"
        FROM "pl_diagram"."ModelCoordSet"
    ''')

    op.execute('''
        INSERT INTO "pl_diagram"."ModelCoordSetGridSize"
        ("coordSetId", "key", "min", "max", "xGrid", "yGrid")
        SELECT "id", 3 as "key", 0.5 as "min", 1000.0 as "max", 1000 as "xGrid", 1000 as "yGrid"
        FROM "pl_diagram"."ModelCoordSet"
    ''')


def downgrade():
    raise NotImplementedError("Downgrade is not implemented")



import logging
from typing import List

from peek_plugin_base.PeekVortexUtil import peekServerName, peekClientName
from peek_plugin_base.storage.DbConnection import DbSessionCreator
from peek_plugin_diagram._private.PluginNames import diagramFilt
from peek_plugin_diagram._private.storage.LocationIndex import LocationIndexCompiled
from peek_plugin_diagram._private.storage.ModelSet import ModelSet
from peek_plugin_diagram._private.tuples.location_index.EncodedLocationIndexTuple import \
    EncodedLocationIndexTuple
from peek_plugin_diagram._private.tuples.location_index.LocationIndexTuple import LocationIndexTuple
from vortex.rpc.RPC import vortexRPC

logger = logging.getLogger(__name__)


class ClientLocationIndexLoaderRpc:
    def __init__(self, dbSessionCreator: DbSessionCreator):
        self._dbSessionCreator = dbSessionCreator

    def makeHandlers(self):
        """ Make Handlers

        In this method we start all the RPC handlers
        start() returns an instance of it's self so we can simply yield the result
        of the start method.

        """

        yield self.loadLocationIndexes.start(funcSelf=self)
        logger.debug("RPCs started")

    # -------------
    @vortexRPC(peekServerName, acceptOnlyFromVortex=peekClientName, timeoutSeconds=60,
               additionalFilt=diagramFilt, deferToThread=True)
    def loadLocationIndexes(self, offset: int, count: int) -> List[EncodedLocationIndexTuple]:
        """ Update Page Loader Status

        Tell the server of the latest status of the loader

        """
        session = self._dbSessionCreator()
        try:
            ormLocationIndexes = (session
                .query(LocationIndexCompiled.indexBucket,
                       LocationIndexCompiled.blobData,
                       LocationIndexCompiled.lastUpdate,
                       ModelSet.key)
                .join(ModelSet)
                .order_by(LocationIndexCompiled.id)
                .offset(offset)
                .limit(count)
                .yield_per(200))

            gridTuples: List[LocationIndexTuple] = []
            for obj in ormLocationIndexes:
                gridTuples.append(
                    EncodedLocationIndexTuple(indexBucket=obj.indexBucket,
                                              encodedLocationIndexTuple=obj.blobData,
                                              lastUpdate=obj.lastUpdate,
                                              modelSetKey=obj.key)
                )

            return gridTuples

        finally:
            session.close()



import logging
from typing import List, Optional

from peek_plugin_diagram._private.client.controller.BranchIndexCacheController import \
    clientBranchIndexUpdateFromServerFilt
from twisted.internet.defer import Deferred

from peek_plugin_base.PeekVortexUtil import peekClientName
from peek_plugin_diagram._private.storage.branch.BranchIndexEncodedChunk import BranchIndexEncodedChunk
from vortex.DeferUtil import vortexLogFailure, deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.VortexFactory import VortexFactory, NoVortexException

logger = logging.getLogger(__name__)


class BranchIndexChunkUpdateHandler:
    """ Client BranchIndex Update Controller

    This controller handles sending updates the the client.

    It uses lower level Vortex API

    It does the following a broadcast to all clients:

    1) Sends grid updates to the clients

    2) Sends Lookup updates to the clients

    """

    def __init__(self, dbSessionCreator):
        self._dbSessionCreator = dbSessionCreator

    def shutdown(self):
        pass

    def sendChunks(self, chunkKeys: List[str]) -> None:
        """ Send Location Indexes

        Send grid updates to the client services

        :param chunkKeys: A list of object buckets that have been updated
        :returns: Nobranch
        """

        if not chunkKeys:
            return

        if peekClientName not in VortexFactory.getRemoteVortexName():
            logger.debug("No clients are online to send the branch chunk to, %s", chunkKeys)
            return

        def send(vortexMsg: bytes):
            if vortexMsg:
                VortexFactory.sendVortexMsg(
                    vortexMsg, destVortexName=peekClientName
                )

        d: Deferred = self._loadChunks(chunkKeys)
        d.addCallback(send)
        d.addErrback(self._sendErrback, chunkKeys)

    def _sendErrback(self, failure, chunkKeys):

        if failure.check(NoVortexException):
            logger.debug(
                "No clients are online to send the branch chunk to, %s", chunkKeys)
            return

        vortexLogFailure(failure, logger)

    @deferToThreadWrapWithLogger(logger)
    def _loadChunks(self, chunkKeys: List[str]) -> Optional[bytes]:

        session = self._dbSessionCreator()
        try:
            results = list(
                session.query(BranchIndexEncodedChunk)
                    .filter(BranchIndexEncodedChunk.chunkKey.in_(chunkKeys))
            )

            if not results:
                return None

            return (
                Payload(filt=clientBranchIndexUpdateFromServerFilt, tuples=results)
                    .makePayloadEnvelope().toVortexMsg()
            )

        finally:
            session.close()



"""Added default group ptr to coord set

Peek Plugin Database Migration Script

Revision ID: aa76b059b81e
Revises: d66cda728b26
Create Date: 2019-05-26 16:00:44.664243

"""

# revision identifiers, used by Alembic.
revision = 'aa76b059b81e'
down_revision = 'f4dce3e782ec'
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.add_column('DispGroupPointer',
                  sa.Column('targetDispGroupName', sa.String(), server_default='0', nullable=True),
                  schema='pl_diagram')

    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('ModelCoordSet',
                  sa.Column('editDefaultEdgeCoordSetId', sa.Integer(), nullable=True),
                  schema='pl_diagram')
    op.add_column('ModelCoordSet',
                  sa.Column('editDefaultEdgeGroupName', sa.String(), nullable=True),
                  schema='pl_diagram')
    op.add_column('ModelCoordSet',
                  sa.Column('editDefaultVertexCoordSetId', sa.Integer(), nullable=True),
                  schema='pl_diagram')
    op.add_column('ModelCoordSet',
                  sa.Column('editDefaultVertexGroupName', sa.String(), nullable=True),
                  schema='pl_diagram')
    op.create_index('idxCoordModel_editDefaultEdgeCoordSetId', 'ModelCoordSet',
                    ['editDefaultEdgeCoordSetId'], unique=False, schema='pl_diagram')
    op.create_index('idxCoordModel_editDefaultVertexCoordSetId', 'ModelCoordSet',
                    ['editDefaultVertexCoordSetId'], unique=False, schema='pl_diagram')
    op.create_foreign_key('ModelCoordSet_editDefaultEdgeCoordSetId_fkey', 'ModelCoordSet',
                          'ModelCoordSet', ['editDefaultEdgeCoordSetId'], ['id'],
                          source_schema='pl_diagram', referent_schema='pl_diagram')
    op.create_foreign_key('ModelCoordSet_editDefaultVertexCoordSetId_fkey',
                          'ModelCoordSet', 'ModelCoordSet',
                          ['editDefaultVertexCoordSetId'], ['id'],
                          source_schema='pl_diagram', referent_schema='pl_diagram')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint('ModelCoordSet_editDefaultVertexCoordSetId_fkey', 'ModelCoordSet',
                       schema='pl_diagram', type_='foreignkey')
    op.drop_constraint('ModelCoordSet_editDefaultEdgeCoordSetId_fkey', 'ModelCoordSet',
                       schema='pl_diagram', type_='foreignkey')
    op.drop_index('idxCoordModel_editDefaultVertexCoordSetId', table_name='ModelCoordSet',
                  schema='pl_diagram')
    op.drop_index('idxCoordModel_editDefaultEdgeCoordSetId', table_name='ModelCoordSet',
                  schema='pl_diagram')
    op.drop_column('ModelCoordSet', 'editDefaultVertexGroupName', schema='pl_diagram')
    op.drop_column('ModelCoordSet', 'editDefaultVertexCoordSetId', schema='pl_diagram')
    op.drop_column('ModelCoordSet', 'editDefaultEdgeGroupName', schema='pl_diagram')
    op.drop_column('ModelCoordSet', 'editDefaultEdgeCoordSetId', schema='pl_diagram')
    op.drop_column('DispGroupPointer', 'name', schema='pl_diagram')
    # ### end Alembic commands ###



import logging
from typing import List, Optional

from sqlalchemy.orm import joinedload
from twisted.internet.defer import Deferred

from peek_plugin_base.PeekVortexUtil import peekClientName
from peek_plugin_base.storage.StorageUtil import makeOrmValuesSubqueryCondition
from peek_plugin_diagram._private.client.controller.LocationIndexCacheController import \
    clientLocationIndexUpdateFromServerFilt
from peek_plugin_diagram._private.storage.LocationIndex import LocationIndexCompiled
from peek_plugin_diagram._private.tuples.location_index.EncodedLocationIndexTuple import \
    EncodedLocationIndexTuple
from peek_plugin_diagram._private.tuples.location_index.LocationIndexTuple import \
    LocationIndexTuple
from vortex.DeferUtil import vortexLogFailure, deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.VortexFactory import VortexFactory, NoVortexException

logger = logging.getLogger(__name__)


class ClientLocationIndexUpdateHandler:
    """ Client Update Controller

    This controller handles sending updates the the client.

    It uses lower level Vortex API

    It does the following a broadcast to all clients:

    1) Sends grid updates to the clients

    2) Sends Lookup updates to the clients

    """

    def __init__(self, dbSessionCreator):
        self._dbSessionCreator = dbSessionCreator

    def shutdown(self):
        pass

    def sendLocationIndexes(self, indexBuckets: List[str]) -> None:
        """ Send Location Indexes

        Send grid updates to the client services

        :param indexBuckets: A list of location index buckets that have been updated
        :returns: Nothing
        """

        if not indexBuckets:
            return

        if peekClientName not in VortexFactory.getRemoteVortexName():
            logger.debug("No clients are online to send the location index to, %s",
                         indexBuckets)
            return

        def send(vortexMsg: bytes):
            if vortexMsg:
                VortexFactory.sendVortexMsg(
                    vortexMsg, destVortexName=peekClientName
                )

        d: Deferred = self._serialiseLocationIndexes(indexBuckets)
        d.addCallback(send)
        d.addErrback(self._sendErrback, indexBuckets)

    def _sendErrback(self, failure, indexBuckets):

        if failure.check(NoVortexException):
            logger.debug(
                "No clients are online to send the location index to, %s", indexBuckets)
            return

        vortexLogFailure(failure, logger)

    @deferToThreadWrapWithLogger(logger)
    def _serialiseLocationIndexes(self, indexBuckets: List[str]) -> Optional[bytes]:
        session = self._dbSessionCreator()
        try:
            ormObjs = (
                session.query(LocationIndexCompiled)
                    .options(joinedload(LocationIndexCompiled.modelSet))
                    .filter(makeOrmValuesSubqueryCondition(
                    session, LocationIndexCompiled.indexBucket, indexBuckets
                ))
                    .yield_per(200)
            )

            locationIndexTuples: List[LocationIndexTuple] = []
            for ormObj in ormObjs:
                locationIndexTuples.append(
                    EncodedLocationIndexTuple(
                        modelSetKey=ormObj.modelSet.key,
                        indexBucket=ormObj.indexBucket,
                        encodedLocationIndexTuple=ormObj.blobData,
                        lastUpdate=ormObj.lastUpdate
                    )
                )

            if not locationIndexTuples:
                return None

            return Payload(
                filt=clientLocationIndexUpdateFromServerFilt, tuples=locationIndexTuples
            ).makePayloadEnvelope().toVortexMsg()

        finally:
            session.close()



from typing import Dict

from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from vortex.Tuple import addTupleType, TupleField, Tuple


@addTupleType
class BranchIndexUpdateDateTuple(Tuple):
    """ BranchIndex Object Update Date Tuple

    This tuple represents the state of the chunks in the cache.
    Each chunkKey has a lastUpdateDate as a string, this is used for offline caching
    all the chunks.
    """
    __tupleType__ = diagramTuplePrefix + "BranchIndexUpdateDateTuple"

    # Improve performance of the JSON serialisation
    __rawJonableFields__ = ('initialLoadComplete', 'updateDateByChunkKey')

    initialLoadComplete: bool = TupleField()
    updateDateByChunkKey: Dict[str, str] = TupleField({})



from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

from peek_plugin_diagram._private.PluginNames import diagramFilt
from peek_plugin_diagram._private.PluginNames import diagramObservableName
from peek_plugin_diagram._private.server.controller.BranchLiveEditController import \
    BranchLiveEditController
from peek_plugin_diagram._private.server.controller.StatusController import \
    StatusController
from peek_plugin_diagram._private.server.tuple_providers.BranchKeyToIdMapProvider import \
    BranchKeyToIdMapTupleProvider
from peek_plugin_diagram._private.server.tuple_providers.BranchLiveEditTupleProvider import \
    BranchLiveEditTupleProvider
from peek_plugin_diagram._private.server.tuple_providers.DiagramLoaderStatusTupleProvider import \
    DiagramLoaderStatusTupleProvider
from peek_plugin_diagram._private.server.tuple_providers.ServerCoordSetTupleProvider import \
    ServerCoordSetTupleProvider
from peek_plugin_diagram._private.server.tuple_providers.ServerLookupTupleProvider import \
    ServerLookupTupleProvider
from peek_plugin_diagram._private.server.tuple_providers.ServerModelSetTupleProvider import \
    ServerModelSetTupleProvider
from peek_plugin_diagram._private.storage.Display import DispLevel, DispLayer, DispColor, \
    DispLineStyle, DispTextStyle
from peek_plugin_diagram._private.storage.ModelSet import ModelCoordSet, ModelSet
from peek_plugin_diagram._private.tuples.DiagramImporterStatusTuple import \
    DiagramImporterStatusTuple
from peek_plugin_diagram._private.tuples.branch.BranchKeyToIdMapTuple import \
    BranchKeyToIdMapTuple
from peek_plugin_diagram._private.tuples.branch.BranchLiveEditTuple import \
    BranchLiveEditTuple


def makeTupleDataObservableHandler(ormSessionCreator,
                                   statusController: StatusController,
                                   branchLiveEditController: BranchLiveEditController):
    """" Make Tuple Data Observable Handler

    This method creates the observable object, registers the tuple providers and then
    returns it.

    :param branchLiveEditController:
    :param ormSessionCreator: A callable that returns an SQLAlchemy session
    :param statusController: The status controller
    :return: An instance of :code:`TupleDataObservableHandler`

    """
    tupleObservable = TupleDataObservableHandler(
        observableName=diagramObservableName,
        additionalFilt=diagramFilt)

    # Register TupleProviders here
    tupleObservable.addTupleProvider(DiagramImporterStatusTuple.tupleName(),
                                     DiagramLoaderStatusTupleProvider(statusController))

    # Register TupleProviders here
    tupleObservable.addTupleProvider(ModelSet.tupleName(),
                                     ServerModelSetTupleProvider(ormSessionCreator))

    # Register TupleProviders here
    tupleObservable.addTupleProvider(ModelCoordSet.tupleName(),
                                     ServerCoordSetTupleProvider(ormSessionCreator))

    # Register TupleProviders here
    tupleObservable.addTupleProvider(BranchKeyToIdMapTuple.tupleName(),
                                     BranchKeyToIdMapTupleProvider(ormSessionCreator))

    # Register TupleProviders here
    lookupTupleProvider = ServerLookupTupleProvider(ormSessionCreator)
    tupleObservable.addTupleProvider(DispLevel.tupleName(), lookupTupleProvider)
    tupleObservable.addTupleProvider(DispLayer.tupleName(), lookupTupleProvider)
    tupleObservable.addTupleProvider(DispColor.tupleName(), lookupTupleProvider)
    tupleObservable.addTupleProvider(DispLineStyle.tupleName(), lookupTupleProvider)
    tupleObservable.addTupleProvider(DispTextStyle.tupleName(), lookupTupleProvider)

    tupleObservable.addTupleProvider(
        BranchLiveEditTuple.tupleName(),
        BranchLiveEditTupleProvider(branchLiveEditController)
    )

    return tupleObservable



import logging
import pytz
from datetime import datetime
from twisted.internet import reactor
from twisted.internet.defer import Deferred
from vortex.TupleAction import TupleActionABC
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleActionProcessor import TupleActionProcessorDelegateABC
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

from peek_plugin_diagram._private.tuples.DiagramImporterStatusTuple import \
    DiagramImporterStatusTuple

logger = logging.getLogger(__name__)


class StatusController(TupleActionProcessorDelegateABC):
    NOTIFY_PERIOD = 2.0

    def __init__(self):
        self._status = DiagramImporterStatusTuple()
        self._tupleObservable = None
        self._notifyPending = False
        self._lastNotifyDatetime = datetime.now(pytz.utc)

    def setTupleObservable(self, tupleObserver: TupleDataObservableHandler):
        self._tupleObserver = tupleObserver

    def shutdown(self):
        self._tupleObserver = None

    def processTupleAction(self, tupleAction: TupleActionABC) -> Deferred:
        # if isinstance(tupleAction, AddIntValueActionTuple):
        #     return self._processAddIntValue(tupleAction)

        raise NotImplementedError(tupleAction.tupleName())

    @property
    def status(self):
        return self._status

    # ---------------
    # Display Compiler Methods

    def setDisplayCompilerStatus(self, state: bool, queueSize: int):
        self._status.displayCompilerQueueStatus = state
        self._status.displayCompilerQueueSize = queueSize
        self._notify()

    def addToDisplayCompilerTotal(self, delta: int):
        self._status.displayCompilerProcessedTotal += delta
        self._notify()

    def setDisplayCompilerError(self, error: str):
        self._status.displayCompilerLastError = error
        self._notify()

    # ---------------
    # Grid Compiler Methods

    def setGridCompilerStatus(self, state: bool, queueSize: int):
        self._status.gridCompilerQueueStatus = state
        self._status.gridCompilerQueueSize = queueSize
        self._notify()

    def addToGridCompilerTotal(self, delta: int):
        self._status.gridCompilerProcessedTotal += delta
        self._notify()

    def setGridCompilerError(self, error: str):
        self._status.gridCompilerLastError = error
        self._notify()

    # ---------------
    # Disp Key Index Compiler Methods

    def setLocationIndexCompilerStatus(self, state: bool, queueSize: int):
        self._status.locationIndexCompilerQueueStatus = state
        self._status.locationIndexCompilerQueueSize = queueSize
        self._notify()

    def addToLocationIndexCompilerTotal(self, delta: int):
        self._status.locationIndexCompilerProcessedTotal += delta
        self._notify()

    def setLocationIndexCompilerError(self, error: str):
        self._status.locationIndexCompilerLastError = error
        self._notify()

    # ---------------
    # Disp Key Index Compiler Methods

    def setBranchIndexCompilerStatus(self, state: bool, queueSize: int):
        self._status.branchIndexCompilerQueueStatus = state
        self._status.branchIndexCompilerQueueSize = queueSize
        self._notify()

    def addToBranchIndexCompilerTotal(self, delta: int):
        self._status.branchIndexCompilerProcessedTotal += delta
        self._notify()

    def setBranchIndexCompilerError(self, error: str):
        self._status.branchIndexCompilerLastError = error
        self._notify()

    # ---------------
    # Notify Methods

    def _notify(self):
        if self._notifyPending:
            return

        self._notifyPending = True

        deltaSeconds = (datetime.now(pytz.utc) - self._lastNotifyDatetime).seconds
        if deltaSeconds >= self.NOTIFY_PERIOD:
            self._sendNotify()
        else:
            reactor.callLater(self.NOTIFY_PERIOD - deltaSeconds, self._sendNotify)

    def _sendNotify(self):
        self._notifyPending = False
        self._lastNotifyDatetime = datetime.now(pytz.utc)
        self._tupleObserver.notifyOfTupleUpdate(
            TupleSelector(DiagramImporterStatusTuple.tupleType(), {})
        )



import logging
from typing import Dict, List

from peek_plugin_diagram._private.PluginNames import diagramFilt
from peek_plugin_diagram._private.server.client_handlers.ClientGridLoaderRpc import \
    ClientGridLoaderRpc
from peek_plugin_diagram._private.tuples.grid.EncodedGridTuple import EncodedGridTuple
from peek_plugin_diagram._private.tuples.grid.GridTuple import GridTuple
from twisted.internet.defer import inlineCallbacks, Deferred
from vortex.DeferUtil import vortexLogFailure
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope

logger = logging.getLogger(__name__)

clientGridUpdateFromServerFilt = dict(key="clientGridUpdateFromServer")
clientGridUpdateFromServerFilt.update(diagramFilt)

clientCoordSetUpdateFromServerFilt = dict(key="clientCoordSetUpdateFromServer")
clientCoordSetUpdateFromServerFilt.update(diagramFilt)


class GridCacheController:
    """ Grid Cache Controller

    The grid cache controller stores all the grids in memory, allowing fast access from
    the mobile and desktop devices.

    NOTE: The grid set endpoint triggers a reload, this is in the case when grid sets
    are enable or disabled. Perhaps the system should just be restarted instead.

    """

    #: This stores the cache of grid data for the clients
    _gridCache: Dict[str, GridTuple] = None

    LOAD_CHUNK = 75
    LOAD_PARALLELISM = 6

    def __init__(self, clientId: str):
        self._clientId = clientId
        self._gridCacheHandler = None
        self._gridCache: Dict[str, EncodedGridTuple] = {}

        self._gridEndpoint = PayloadEndpoint(clientGridUpdateFromServerFilt,
                                             self._processGridPayload)

        self._coordSetEndpoint = PayloadEndpoint(clientCoordSetUpdateFromServerFilt,
                                                 self._processCoordSetPayload)

    def setGridCacheHandler(self, gridCacheHandler):
        self._gridCacheHandler = gridCacheHandler

    @inlineCallbacks
    def start(self):
        yield self.reloadCache()

    def shutdown(self):
        self._tupleObservable = None

        self._gridEndpoint.shutdown()
        self._gridEndpoint = None

        self._coordSetEndpoint.shutdown()
        self._coordSetEndpoint = None

        self._gridCache = {}
        # self._cachedGridCoordSetIds = set()

    def reloadCache(self):
        self._gridCache = {}

        d = Deferred()

        class C:
            inProgress = self.LOAD_PARALLELISM

        def nextChunk():
            offset = 0
            while True:
                yield offset
                offset += self.LOAD_CHUNK

        nextChunkGen = nextChunk()

        def callback(gridTuples):
            if not gridTuples:
                C.inProgress -= 1
                if not C.inProgress:
                    d.callback(True)
                return

            self._loadGridIntoCache(gridTuples)
            load()

        def load():
            offset = next(nextChunkGen)
            logger.info("Loading grids %s to %s" %
                        (offset, offset + self.LOAD_CHUNK))
            d = ClientGridLoaderRpc.loadGrids(offset, self.LOAD_CHUNK)
            d.addCallback(callback)
            d.addErrback(vortexLogFailure, logger, consumeError=True)

        for _ in range(self.LOAD_PARALLELISM):
            load()

        return d

    @inlineCallbacks
    def _processGridPayload(self, payloadEnvelope: PayloadEnvelope, **kwargs):
        payload = yield payloadEnvelope.decodePayloadDefer()
        gridTuples: List[EncodedGridTuple] = payload.tuples
        return self._loadGridIntoCache(gridTuples)

    def _processCoordSetPayload(self, payloadEnvelope: PayloadEnvelope, **kwargs):
        d: Deferred = self.reloadCache()
        d.addErrback(vortexLogFailure, logger, consumeError=True)

    def _loadGridIntoCache(self, encodedGridTuples: List[EncodedGridTuple]):
        gridKeyUpdates: List[str] = []

        for t in encodedGridTuples:
            if (not t.gridKey in self._gridCache or
                    self._gridCache[t.gridKey].lastUpdate != t.lastUpdate):
                self._gridCache[t.gridKey] = t
                gridKeyUpdates.append(t.gridKey)

        logger.debug("Received %s grids from server,"
                     " %s had changed ",
                     len(encodedGridTuples), len(gridKeyUpdates))

        self._gridCacheHandler.notifyOfGridUpdate(gridKeyUpdates)

    def grid(self, gridKey) -> EncodedGridTuple:
        return self._gridCache.get(gridKey)

    def gridKeyList(self) -> List[str]:
        return list(self._gridCache.keys())

    def gridDatesByKey(self):
        return {g.gridKey: g.lastUpdate for g in self._gridCache.values()}



"""timezone aware

Peek Plugin Database Migration Script

Revision ID: cea17a52b9ae
Revises: 34dd05f474df
Create Date: 2017-12-31 14:29:21.485753

"""

# revision identifiers, used by Alembic.
revision = 'cea17a52b9ae'
down_revision = '34dd05f474df'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():

    # pl_diagram."DispBase"
    op.execute('''ALTER TABLE pl_diagram."DispBase"
                  ALTER COLUMN "importUpdateDate" TYPE TIMESTAMP WITH TIME ZONE
                  USING "importUpdateDate" AT TIME ZONE 'UTC'
               ''')


def downgrade():
    pass


"""Removed Node and Conn

Peek Plugin Database Migration Script

Revision ID: 2c5f47322bda
Revises: d6e25d8ba156
Create Date: 2017-08-24 12:58:59.016532

"""

# revision identifiers, used by Alembic.
revision = '2c5f47322bda'
down_revision = 'd6e25d8ba156'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('DispPolylineConn', schema='pl_diagram')
    op.drop_table('DispGroupPointerNode', schema='pl_diagram')
    op.drop_table('ModelConn', schema='pl_diagram')
    op.drop_table('ModelNode', schema='pl_diagram')
    op.drop_table('DispAction', schema='pl_diagram')
    op.drop_table('ModelNodeType', schema='pl_diagram')
    op.drop_table('ModelConnType', schema='pl_diagram')

    op.add_column('DispBase', sa.Column('key', sa.String(length=50), nullable=True), schema='pl_diagram')
    op.add_column('DispBase', sa.Column('selectable', sa.Boolean(), server_default='0', nullable=False), schema='pl_diagram')

    op.add_column('ModelSet', sa.Column('key', sa.String(length=50), nullable=True), schema='pl_diagram')
    op.execute(''' UPDATE "pl_diagram"."ModelSet" SET "key" = "name" ''')
    op.alter_column('ModelSet', 'key', type_=sa.String(length=50), nullable=False, schema='pl_diagram')

    # ### end Alembic commands ###




def downgrade():
    raise NotImplementedError("Downgrade is not implemented")